from collections.abc import Mapping
from typing import Literal

# Locale Identifiers - https://gist.github.com/ndbroadbent/588fefab8e0f1b459fcec8181b41b39c
type Locale = Literal[
    "af",
    "af-NA",
    "af-ZA",
    "agq",
    "agq-CM",
    "ak",
    "ak-GH",
    "am",
    "am-ET",
    "ar",
    "ar-001",
    "ar-AE",
    "ar-BH",
    "ar-DJ",
    "ar-DZ",
    "ar-EG",
    "ar-EH",
    "ar-ER",
    "ar-IL",
    "ar-IQ",
    "ar-JO",
    "ar-KM",
    "ar-KW",
    "ar-LB",
    "ar-LY",
    "ar-MA",
    "ar-MR",
    "ar-OM",
    "ar-PS",
    "ar-QA",
    "ar-SA",
    "ar-SD",
    "ar-SO",
    "ar-SS",
    "ar-SY",
    "ar-TD",
    "ar-TN",
    "ar-YE",
    "as",
    "as-IN",
    "asa",
    "asa-TZ",
    "az",
    "az-Cyrl",
    "az-Cyrl-AZ",
    "az-Latn",
    "az-Latn-AZ",
    "bas",
    "bas-CM",
    "be",
    "be-BY",
    "bem",
    "bem-ZM",
    "bez",
    "bez-TZ",
    "bg",
    "bg-BG",
    "bm",
    "bm-Latn",
    "bm-Latn-ML",
    "bn",
    "bn-BD",
    "bn-IN",
    "bo",
    "bo-CN",
    "bo-IN",
    "br",
    "br-FR",
    "brx",
    "brx-IN",
    "bs",
    "bs-Cyrl",
    "bs-Cyrl-BA",
    "bs-Latn",
    "bs-Latn-BA",
    "ca",
    "ca-AD",
    "ca-ES",
    "ca-FR",
    "ca-IT",
    "cgg",
    "cgg-UG",
    "chr",
    "chr-US",
    "ckb",
    "ckb-IQ",
    "ckb-IR",
    "cs",
    "cs-CZ",
    "cy",
    "cy-GB",
    "da",
    "da-DK",
    "da-GL",
    "dav",
    "dav-KE",
    "de",
    "de-AT",
    "de-BE",
    "de-CH",
    "de-DE",
    "de-LI",
    "de-LU",
    "dje",
    "dje-NE",
    "dsb",
    "dsb-DE",
    "dua",
    "dua-CM",
    "dyo",
    "dyo-SN",
    "dz",
    "dz-BT",
    "ebu",
    "ebu-KE",
    "ee",
    "ee-GH",
    "ee-TG",
    "el",
    "el-CY",
    "el-GR",
    "en",
    "en-001",
    "en-150",
    "en-AD",
    "en-AG",
    "en-AI",
    "en-AL",
    "en-AS",
    "en-AT",
    "en-AU",
    "en-BA",
    "en-BB",
    "en-BE",
    "en-BM",
    "en-BS",
    "en-BW",
    "en-BZ",
    "en-CA",
    "en-CC",
    "en-CH",
    "en-CK",
    "en-CM",
    "en-CX",
    "en-CY",
    "en-CZ",
    "en-DE",
    "en-DG",
    "en-DK",
    "en-DM",
    "en-EE",
    "en-ER",
    "en-ES",
    "en-FI",
    "en-FJ",
    "en-FK",
    "en-FM",
    "en-FR",
    "en-GB",
    "en-GD",
    "en-GG",
    "en-GH",
    "en-GI",
    "en-GM",
    "en-GR",
    "en-GU",
    "en-GY",
    "en-HK",
    "en-HR",
    "en-HU",
    "en-IE",
    "en-IL",
    "en-IM",
    "en-IN",
    "en-IO",
    "en-IS",
    "en-IT",
    "en-JE",
    "en-JM",
    "en-KE",
    "en-KI",
    "en-KN",
    "en-KY",
    "en-LC",
    "en-LR",
    "en-LS",
    "en-LT",
    "en-LU",
    "en-LV",
    "en-ME",
    "en-MG",
    "en-MH",
    "en-MO",
    "en-MP",
    "en-MS",
    "en-MT",
    "en-MU",
    "en-MW",
    "en-MY",
    "en-NA",
    "en-NF",
    "en-NG",
    "en-NL",
    "en-NO",
    "en-NR",
    "en-NU",
    "en-NZ",
    "en-PG",
    "en-PH",
    "en-PK",
    "en-PL",
    "en-PN",
    "en-PR",
    "en-PT",
    "en-PW",
    "en-RO",
    "en-RU",
    "en-RW",
    "en-SB",
    "en-SC",
    "en-SD",
    "en-SE",
    "en-SG",
    "en-SH",
    "en-SI",
    "en-SK",
    "en-SL",
    "en-SS",
    "en-SX",
    "en-SZ",
    "en-TC",
    "en-TK",
    "en-TO",
    "en-TR",
    "en-TT",
    "en-TV",
    "en-TZ",
    "en-UG",
    "en-UM",
    "en-US",
    "en-US-POSIX",
    "en-VC",
    "en-VG",
    "en-VI",
    "en-VU",
    "en-WS",
    "en-ZA",
    "en-ZM",
    "en-ZW",
    "eo",
    "es",
    "es-419",
    "es-AR",
    "es-BO",
    "es-CL",
    "es-CO",
    "es-CR",
    "es-CU",
    "es-DO",
    "es-EA",
    "es-EC",
    "es-ES",
    "es-GQ",
    "es-GT",
    "es-HN",
    "es-IC",
    "es-MX",
    "es-NI",
    "es-PA",
    "es-PE",
    "es-PH",
    "es-PR",
    "es-PY",
    "es-SV",
    "es-US",
    "es-UY",
    "es-VE",
    "et",
    "et-EE",
    "eu",
    "eu-ES",
    "ewo",
    "ewo-CM",
    "fa",
    "fa-AF",
    "fa-IR",
    "ff",
    "ff-CM",
    "ff-GN",
    "ff-MR",
    "ff-SN",
    "fi",
    "fi-FI",
    "fil",
    "fil-PH",
    "fo",
    "fo-FO",
    "fr",
    "fr-BE",
    "fr-BF",
    "fr-BI",
    "fr-BJ",
    "fr-BL",
    "fr-CA",
    "fr-CD",
    "fr-CF",
    "fr-CG",
    "fr-CH",
    "fr-CI",
    "fr-CM",
    "fr-DJ",
    "fr-DZ",
    "fr-FR",
    "fr-GA",
    "fr-GF",
    "fr-GN",
    "fr-GP",
    "fr-GQ",
    "fr-HT",
    "fr-KM",
    "fr-LU",
    "fr-MA",
    "fr-MC",
    "fr-MF",
    "fr-MG",
    "fr-ML",
    "fr-MQ",
    "fr-MR",
    "fr-MU",
    "fr-NC",
    "fr-NE",
    "fr-PF",
    "fr-PM",
    "fr-RE",
    "fr-RW",
    "fr-SC",
    "fr-SN",
    "fr-SY",
    "fr-TD",
    "fr-TG",
    "fr-TN",
    "fr-VU",
    "fr-WF",
    "fr-YT",
    "fur",
    "fur-IT",
    "fy",
    "fy-NL",
    "ga",
    "ga-IE",
    "gd",
    "gd-GB",
    "gl",
    "gl-ES",
    "gsw",
    "gsw-CH",
    "gsw-FR",
    "gsw-LI",
    "gu",
    "gu-IN",
    "guz",
    "guz-KE",
    "gv",
    "gv-IM",
    "ha",
    "ha-Latn",
    "ha-Latn-GH",
    "ha-Latn-NE",
    "ha-Latn-NG",
    "haw",
    "haw-US",
    "he",
    "he-IL",
    "hi",
    "hi-IN",
    "hr",
    "hr-BA",
    "hr-HR",
    "hsb",
    "hsb-DE",
    "hu",
    "hu-HU",
    "hy",
    "hy-AM",
    "id",
    "id-ID",
    "ig",
    "ig-NG",
    "ii",
    "ii-CN",
    "is",
    "is-IS",
    "it",
    "it-CH",
    "it-IT",
    "it-SM",
    "iu",
    "iu-Cans",
    "iu-Cans-CA",
    "ja",
    "ja-JP",
    "jgo",
    "jgo-CM",
    "jmc",
    "jmc-TZ",
    "ka",
    "ka-GE",
    "kab",
    "kab-DZ",
    "kam",
    "kam-KE",
    "kde",
    "kde-TZ",
    "kea",
    "kea-CV",
    "khq",
    "khq-ML",
    "ki",
    "ki-KE",
    "kk",
    "kk-Cyrl",
    "kk-Cyrl-KZ",
    "kkj",
    "kkj-CM",
    "kl",
    "kl-GL",
    "kln",
    "kln-KE",
    "km",
    "km-KH",
    "kn",
    "kn-IN",
    "ko",
    "ko-KP",
    "ko-KR",
    "kok",
    "kok-IN",
    "ks",
    "ks-Arab",
    "ks-Arab-IN",
    "ksb",
    "ksb-TZ",
    "ksf",
    "ksf-CM",
    "ksh",
    "ksh-DE",
    "kw",
    "kw-GB",
    "ky",
    "ky-Cyrl",
    "ky-Cyrl-KG",
    "lag",
    "lag-TZ",
    "lb",
    "lb-LU",
    "lg",
    "lg-UG",
    "lkt",
    "lkt-US",
    "ln",
    "ln-AO",
    "ln-CD",
    "ln-CF",
    "ln-CG",
    "lo",
    "lo-LA",
    "localeIdentifier",
    "lt",
    "lt-LT",
    "lu",
    "lu-CD",
    "luo",
    "luo-KE",
    "luy",
    "luy-KE",
    "lv",
    "lv-LV",
    "mas",
    "mas-KE",
    "mas-TZ",
    "mer",
    "mer-KE",
    "mfe",
    "mfe-MU",
    "mg",
    "mg-MG",
    "mgh",
    "mgh-MZ",
    "mgo",
    "mgo-CM",
    "mk",
    "mk-MK",
    "ml",
    "ml-IN",
    "mn",
    "mn-Cyrl",
    "mn-Cyrl-MN",
    "mr",
    "mr-IN",
    "ms",
    "ms-Arab",
    "ms-Arab-BN",
    "ms-Arab-MY",
    "ms-Latn",
    "ms-Latn-BN",
    "ms-Latn-MY",
    "ms-Latn-SG",
    "mt",
    "mt-MT",
    "mua",
    "mua-CM",
    "my",
    "my-MM",
    "naq",
    "naq-NA",
    "nb",
    "nb-NO",
    "nb-SJ",
    "nd",
    "nd-ZW",
    "ne",
    "ne-IN",
    "ne-NP",
    "nl",
    "nl-AW",
    "nl-BE",
    "nl-BQ",
    "nl-CW",
    "nl-NL",
    "nl-SR",
    "nl-SX",
    "nmg",
    "nmg-CM",
    "nn",
    "nn-NO",
    "nnh",
    "nnh-CM",
    "nus",
    "nus-SD",
    "nyn",
    "nyn-UG",
    "om",
    "om-ET",
    "om-KE",
    "or",
    "or-IN",
    "os",
    "os-GE",
    "os-RU",
    "pa",
    "pa-Arab",
    "pa-Arab-PK",
    "pa-Guru",
    "pa-Guru-IN",
    "pl",
    "pl-PL",
    "ps",
    "ps-AF",
    "pt",
    "pt-AO",
    "pt-BR",
    "pt-CV",
    "pt-GW",
    "pt-MO",
    "pt-MZ",
    "pt-PT",
    "pt-ST",
    "pt-TL",
    "qu",
    "qu-BO",
    "qu-EC",
    "qu-PE",
    "rm",
    "rm-CH",
    "rn",
    "rn-BI",
    "ro",
    "ro-MD",
    "ro-RO",
    "rof",
    "rof-TZ",
    "ru",
    "ru-BY",
    "ru-KG",
    "ru-KZ",
    "ru-MD",
    "ru-RU",
    "ru-UA",
    "rw",
    "rw-RW",
    "rwk",
    "rwk-TZ",
    "sah",
    "sah-RU",
    "saq",
    "saq-KE",
    "sbp",
    "sbp-TZ",
    "se",
    "se-FI",
    "se-NO",
    "se-SE",
    "seh",
    "seh-MZ",
    "ses",
    "ses-ML",
    "sg",
    "sg-CF",
    "shi",
    "shi-Latn",
    "shi-Latn-MA",
    "shi-Tfng",
    "shi-Tfng-MA",
    "si",
    "si-LK",
    "sk",
    "sk-SK",
    "sl",
    "sl-SI",
    "smn",
    "smn-FI",
    "sn",
    "sn-ZW",
    "so",
    "so-DJ",
    "so-ET",
    "so-KE",
    "so-SO",
    "sq",
    "sq-AL",
    "sq-MK",
    "sq-XK",
    "sr",
    "sr-Cyrl",
    "sr-Cyrl-BA",
    "sr-Cyrl-ME",
    "sr-Cyrl-RS",
    "sr-Cyrl-XK",
    "sr-Latn",
    "sr-Latn-BA",
    "sr-Latn-ME",
    "sr-Latn-RS",
    "sr-Latn-XK",
    "sv",
    "sv-AX",
    "sv-FI",
    "sv-SE",
    "sw",
    "sw-CD",
    "sw-KE",
    "sw-TZ",
    "sw-UG",
    "ta",
    "ta-IN",
    "ta-LK",
    "ta-MY",
    "ta-SG",
    "te",
    "te-IN",
    "teo",
    "teo-KE",
    "teo-UG",
    "tg",
    "tg-Cyrl",
    "tg-Cyrl-TJ",
    "th",
    "th-TH",
    "ti",
    "ti-ER",
    "ti-ET",
    "tk",
    "tk-Latn",
    "tk-Latn-TM",
    "to",
    "to-TO",
    "tr",
    "tr-CY",
    "tr-TR",
    "twq",
    "twq-NE",
    "tzm",
    "tzm-Latn",
    "tzm-Latn-MA",
    "ug",
    "ug-Arab",
    "ug-Arab-CN",
    "uk",
    "uk-UA",
    "ur",
    "ur-IN",
    "ur-PK",
    "uz",
    "uz-Arab",
    "uz-Arab-AF",
    "uz-Cyrl",
    "uz-Cyrl-UZ",
    "uz-Latn",
    "uz-Latn-UZ",
    "vai",
    "vai-Latn",
    "vai-Latn-LR",
    "vai-Vaii",
    "vai-Vaii-LR",
    "vi",
    "vi-VN",
    "vun",
    "vun-TZ",
    "wae",
    "wae-CH",
    "xog",
    "xog-UG",
    "yav",
    "yav-CM",
    "yi",
    "yi-001",
    "yo",
    "yo-BJ",
    "yo-NG",
    "zgh",
    "zgh-MA",
    "zh",
    "zh-Hans",
    "zh-Hans-CN",
    "zh-Hans-HK",
    "zh-Hans-MO",
    "zh-Hans-SG",
    "zh-Hant",
    "zh-Hant-HK",
    "zh-Hant-MO",
    "zh-Hant-TW",
    "zu",
    "zu-ZA",
]
LOCALE_NAMES: Mapping[Locale, str] = {
    "af": "Afrikaans",
    "af-NA": "Afrikaans (Namibia)",
    "af-ZA": "Afrikaans (South Africa)",
    "agq": "Aghem",
    "agq-CM": "Aghem (Cameroon)",
    "ak": "Akan",
    "ak-GH": "Akan (Ghana)",
    "am": "Amharic",
    "am-ET": "Amharic (Ethiopia)",
    "ar": "Arabic",
    "ar-001": "Arabic (World)",
    "ar-AE": "Arabic (United Arab Emirates)",
    "ar-BH": "Arabic (Bahrain)",
    "ar-DJ": "Arabic (Djibouti)",
    "ar-DZ": "Arabic (Algeria)",
    "ar-EG": "Arabic (Egypt)",
    "ar-EH": "Arabic (Western Sahara)",
    "ar-ER": "Arabic (Eritrea)",
    "ar-IL": "Arabic (Israel)",
    "ar-IQ": "Arabic (Iraq)",
    "ar-JO": "Arabic (Jordan)",
    "ar-KM": "Arabic (Comoros)",
    "ar-KW": "Arabic (Kuwait)",
    "ar-LB": "Arabic (Lebanon)",
    "ar-LY": "Arabic (Libya)",
    "ar-MA": "Arabic (Morocco)",
    "ar-MR": "Arabic (Mauritania)",
    "ar-OM": "Arabic (Oman)",
    "ar-PS": "Arabic (Palestinian Territories)",
    "ar-QA": "Arabic (Qatar)",
    "ar-SA": "Arabic (Saudi Arabia)",
    "ar-SD": "Arabic (Sudan)",
    "ar-SO": "Arabic (Somalia)",
    "ar-SS": "Arabic (South Sudan)",
    "ar-SY": "Arabic (Syria)",
    "ar-TD": "Arabic (Chad)",
    "ar-TN": "Arabic (Tunisia)",
    "ar-YE": "Arabic (Yemen)",
    "as": "Assamese",
    "as-IN": "Assamese (India)",
    "asa": "Asu",
    "asa-TZ": "Asu (Tanzania)",
    "az": "Azerbaijani",
    "az-Cyrl": "Azerbaijani (Cyrillic)",
    "az-Cyrl-AZ": "Azerbaijani (Cyrillic, Azerbaijan)",
    "az-Latn": "Azerbaijani (Latin)",
    "az-Latn-AZ": "Azerbaijani (Latin, Azerbaijan)",
    "bas": "Basaa",
    "bas-CM": "Basaa (Cameroon)",
    "be": "Belarusian",
    "be-BY": "Belarusian (Belarus)",
    "bem": "Bemba",
    "bem-ZM": "Bemba (Zambia)",
    "bez": "Bena",
    "bez-TZ": "Bena (Tanzania)",
    "bg": "Bulgarian",
    "bg-BG": "Bulgarian (Bulgaria)",
    "bm": "Bambara",
    "bm-Latn": "Bambara (Latin)",
    "bm-Latn-ML": "Bambara (Latin, Mali)",
    "bn": "Bengali",
    "bn-BD": "Bengali (Bangladesh)",
    "bn-IN": "Bengali (India)",
    "bo": "Tibetan",
    "bo-CN": "Tibetan (China)",
    "bo-IN": "Tibetan (India)",
    "br": "Breton",
    "br-FR": "Breton (France)",
    "brx": "Bodo",
    "brx-IN": "Bodo (India)",
    "bs": "Bosnian",
    "bs-Cyrl": "Bosnian (Cyrillic)",
    "bs-Cyrl-BA": "Bosnian (Cyrillic, Bosnia & Herzegovina)",
    "bs-Latn": "Bosnian (Latin)",
    "bs-Latn-BA": "Bosnian (Latin, Bosnia & Herzegovina)",
    "ca": "Catalan",
    "ca-AD": "Catalan (Andorra)",
    "ca-ES": "Catalan (Spain)",
    "ca-FR": "Catalan (France)",
    "ca-IT": "Catalan (Italy)",
    "cgg": "Chiga",
    "cgg-UG": "Chiga (Uganda)",
    "chr": "Cherokee",
    "chr-US": "Cherokee (United States)",
    "ckb": "Central Kurdish",
    "ckb-IQ": "Central Kurdish (Iraq)",
    "ckb-IR": "Central Kurdish (Iran)",
    "cs": "Czech",
    "cs-CZ": "Czech (Czech Republic)",
    "cy": "Welsh",
    "cy-GB": "Welsh (United Kingdom)",
    "da": "Danish",
    "da-DK": "Danish (Denmark)",
    "da-GL": "Danish (Greenland)",
    "dav": "Taita",
    "dav-KE": "Taita (Kenya)",
    "de": "German",
    "de-AT": "German (Austria)",
    "de-BE": "German (Belgium)",
    "de-CH": "German (Switzerland)",
    "de-DE": "German (Germany)",
    "de-LI": "German (Liechtenstein)",
    "de-LU": "German (Luxembourg)",
    "dje": "Zarma",
    "dje-NE": "Zarma (Niger)",
    "dsb": "Lower Sorbian",
    "dsb-DE": "Lower Sorbian (Germany)",
    "dua": "Duala",
    "dua-CM": "Duala (Cameroon)",
    "dyo": "Jola-Fonyi",
    "dyo-SN": "Jola-Fonyi (Senegal)",
    "dz": "Dzongkha",
    "dz-BT": "Dzongkha (Bhutan)",
    "ebu": "Embu",
    "ebu-KE": "Embu (Kenya)",
    "ee": "Ewe",
    "ee-GH": "Ewe (Ghana)",
    "ee-TG": "Ewe (Togo)",
    "el": "Greek",
    "el-CY": "Greek (Cyprus)",
    "el-GR": "Greek (Greece)",
    "en": "English",
    "en-001": "English (World)",
    "en-150": "English (Europe)",
    "en-AD": "English (Andorra)",
    "en-AG": "English (Antigua & Barbuda)",
    "en-AI": "English (Anguilla)",
    "en-AL": "English (Albania)",
    "en-AS": "English (American Samoa)",
    "en-AT": "English (Austria)",
    "en-AU": "English (Australia)",
    "en-BA": "English (Bosnia & Herzegovina)",
    "en-BB": "English (Barbados)",
    "en-BE": "English (Belgium)",
    "en-BM": "English (Bermuda)",
    "en-BS": "English (Bahamas)",
    "en-BW": "English (Botswana)",
    "en-BZ": "English (Belize)",
    "en-CA": "English (Canada)",
    "en-CC": "English (Cocos [Keeling] Islands)",
    "en-CH": "English (Switzerland)",
    "en-CK": "English (Cook Islands)",
    "en-CM": "English (Cameroon)",
    "en-CX": "English (Christmas Island)",
    "en-CY": "English (Cyprus)",
    "en-CZ": "English (Czech Republic)",
    "en-DE": "English (Germany)",
    "en-DG": "English (Diego Garcia)",
    "en-DK": "English (Denmark)",
    "en-DM": "English (Dominica)",
    "en-EE": "English (Estonia)",
    "en-ER": "English (Eritrea)",
    "en-ES": "English (Spain)",
    "en-FI": "English (Finland)",
    "en-FJ": "English (Fiji)",
    "en-FK": "English (Falkland Islands)",
    "en-FM": "English (Micronesia)",
    "en-FR": "English (France)",
    "en-GB": "English (United Kingdom)",
    "en-GD": "English (Grenada)",
    "en-GG": "English (Guernsey)",
    "en-GH": "English (Ghana)",
    "en-GI": "English (Gibraltar)",
    "en-GM": "English (Gambia)",
    "en-GR": "English (Greece)",
    "en-GU": "English (Guam)",
    "en-GY": "English (Guyana)",
    "en-HK": "English (Hong Kong SAR China)",
    "en-HR": "English (Croatia)",
    "en-HU": "English (Hungary)",
    "en-IE": "English (Ireland)",
    "en-IL": "English (Israel)",
    "en-IM": "English (Isle of Man)",
    "en-IN": "English (India)",
    "en-IO": "English (British Indian Ocean Territory)",
    "en-IS": "English (Iceland)",
    "en-IT": "English (Italy)",
    "en-JE": "English (Jersey)",
    "en-JM": "English (Jamaica)",
    "en-KE": "English (Kenya)",
    "en-KI": "English (Kiribati)",
    "en-KN": "English (St. Kitts & Nevis)",
    "en-KY": "English (Cayman Islands)",
    "en-LC": "English (St. Lucia)",
    "en-LR": "English (Liberia)",
    "en-LS": "English (Lesotho)",
    "en-LT": "English (Lithuania)",
    "en-LU": "English (Luxembourg)",
    "en-LV": "English (Latvia)",
    "en-ME": "English (Montenegro)",
    "en-MG": "English (Madagascar)",
    "en-MH": "English (Marshall Islands)",
    "en-MO": "English (Macau SAR China)",
    "en-MP": "English (Northern Mariana Islands)",
    "en-MS": "English (Montserrat)",
    "en-MT": "English (Malta)",
    "en-MU": "English (Mauritius)",
    "en-MW": "English (Malawi)",
    "en-MY": "English (Malaysia)",
    "en-NA": "English (Namibia)",
    "en-NF": "English (Norfolk Island)",
    "en-NG": "English (Nigeria)",
    "en-NL": "English (Netherlands)",
    "en-NO": "English (Norway)",
    "en-NR": "English (Nauru)",
    "en-NU": "English (Niue)",
    "en-NZ": "English (New Zealand)",
    "en-PG": "English (Papua New Guinea)",
    "en-PH": "English (Philippines)",
    "en-PK": "English (Pakistan)",
    "en-PL": "English (Poland)",
    "en-PN": "English (Pitcairn Islands)",
    "en-PR": "English (Puerto Rico)",
    "en-PT": "English (Portugal)",
    "en-PW": "English (Palau)",
    "en-RO": "English (Romania)",
    "en-RU": "English (Russia)",
    "en-RW": "English (Rwanda)",
    "en-SB": "English (Solomon Islands)",
    "en-SC": "English (Seychelles)",
    "en-SD": "English (Sudan)",
    "en-SE": "English (Sweden)",
    "en-SG": "English (Singapore)",
    "en-SH": "English (St. Helena)",
    "en-SI": "English (Slovenia)",
    "en-SK": "English (Slovakia)",
    "en-SL": "English (Sierra Leone)",
    "en-SS": "English (South Sudan)",
    "en-SX": "English (Sint Maarten)",
    "en-SZ": "English (Swaziland)",
    "en-TC": "English (Turks & Caicos Islands)",
    "en-TK": "English (Tokelau)",
    "en-TO": "English (Tonga)",
    "en-TR": "English (Turkey)",
    "en-TT": "English (Trinidad & Tobago)",
    "en-TV": "English (Tuvalu)",
    "en-TZ": "English (Tanzania)",
    "en-UG": "English (Uganda)",
    "en-UM": "English (U.S. Outlying Islands)",
    "en-US": "English (United States)",
    "en-US-POSIX": "English (United States, Computer)",
    "en-VC": "English (St. Vincent & Grenadines)",
    "en-VG": "English (British Virgin Islands)",
    "en-VI": "English (U.S. Virgin Islands)",
    "en-VU": "English (Vanuatu)",
    "en-WS": "English (Samoa)",
    "en-ZA": "English (South Africa)",
    "en-ZM": "English (Zambia)",
    "en-ZW": "English (Zimbabwe)",
    "eo": "Esperanto",
    "es": "Spanish",
    "es-419": "Spanish (Latin America)",
    "es-AR": "Spanish (Argentina)",
    "es-BO": "Spanish (Bolivia)",
    "es-CL": "Spanish (Chile)",
    "es-CO": "Spanish (Colombia)",
    "es-CR": "Spanish (Costa Rica)",
    "es-CU": "Spanish (Cuba)",
    "es-DO": "Spanish (Dominican Republic)",
    "es-EA": "Spanish (Ceuta & Melilla)",
    "es-EC": "Spanish (Ecuador)",
    "es-ES": "Spanish (Spain)",
    "es-GQ": "Spanish (Equatorial Guinea)",
    "es-GT": "Spanish (Guatemala)",
    "es-HN": "Spanish (Honduras)",
    "es-IC": "Spanish (Canary Islands)",
    "es-MX": "Spanish (Mexico)",
    "es-NI": "Spanish (Nicaragua)",
    "es-PA": "Spanish (Panama)",
    "es-PE": "Spanish (Peru)",
    "es-PH": "Spanish (Philippines)",
    "es-PR": "Spanish (Puerto Rico)",
    "es-PY": "Spanish (Paraguay)",
    "es-SV": "Spanish (El Salvador)",
    "es-US": "Spanish (United States)",
    "es-UY": "Spanish (Uruguay)",
    "es-VE": "Spanish (Venezuela)",
    "et": "Estonian",
    "et-EE": "Estonian (Estonia)",
    "eu": "Basque",
    "eu-ES": "Basque (Spain)",
    "ewo": "Ewondo",
    "ewo-CM": "Ewondo (Cameroon)",
    "fa": "Persian",
    "fa-AF": "Persian (Afghanistan)",
    "fa-IR": "Persian (Iran)",
    "ff": "Fulah",
    "ff-CM": "Fulah (Cameroon)",
    "ff-GN": "Fulah (Guinea)",
    "ff-MR": "Fulah (Mauritania)",
    "ff-SN": "Fulah (Senegal)",
    "fi": "Finnish",
    "fi-FI": "Finnish (Finland)",
    "fil": "Filipino",
    "fil-PH": "Filipino (Philippines)",
    "fo": "Faroese",
    "fo-FO": "Faroese (Faroe Islands)",
    "fr": "French",
    "fr-BE": "French (Belgium)",
    "fr-BF": "French (Burkina Faso)",
    "fr-BI": "French (Burundi)",
    "fr-BJ": "French (Benin)",
    "fr-BL": "French (St. Barthélemy)",
    "fr-CA": "French (Canada)",
    "fr-CD": "French (Congo - Kinshasa)",
    "fr-CF": "French (Central African Republic)",
    "fr-CG": "French (Congo - Brazzaville)",
    "fr-CH": "French (Switzerland)",
    "fr-CI": "French (Côte d’Ivoire)",
    "fr-CM": "French (Cameroon)",
    "fr-DJ": "French (Djibouti)",
    "fr-DZ": "French (Algeria)",
    "fr-FR": "French (France)",
    "fr-GA": "French (Gabon)",
    "fr-GF": "French (French Guiana)",
    "fr-GN": "French (Guinea)",
    "fr-GP": "French (Guadeloupe)",
    "fr-GQ": "French (Equatorial Guinea)",
    "fr-HT": "French (Haiti)",
    "fr-KM": "French (Comoros)",
    "fr-LU": "French (Luxembourg)",
    "fr-MA": "French (Morocco)",
    "fr-MC": "French (Monaco)",
    "fr-MF": "French (St. Martin)",
    "fr-MG": "French (Madagascar)",
    "fr-ML": "French (Mali)",
    "fr-MQ": "French (Martinique)",
    "fr-MR": "French (Mauritania)",
    "fr-MU": "French (Mauritius)",
    "fr-NC": "French (New Caledonia)",
    "fr-NE": "French (Niger)",
    "fr-PF": "French (French Polynesia)",
    "fr-PM": "French (St. Pierre & Miquelon)",
    "fr-RE": "French (Réunion)",
    "fr-RW": "French (Rwanda)",
    "fr-SC": "French (Seychelles)",
    "fr-SN": "French (Senegal)",
    "fr-SY": "French (Syria)",
    "fr-TD": "French (Chad)",
    "fr-TG": "French (Togo)",
    "fr-TN": "French (Tunisia)",
    "fr-VU": "French (Vanuatu)",
    "fr-WF": "French (Wallis & Futuna)",
    "fr-YT": "French (Mayotte)",
    "fur": "Friulian",
    "fur-IT": "Friulian (Italy)",
    "fy": "Western Frisian",
    "fy-NL": "Western Frisian (Netherlands)",
    "ga": "Irish",
    "ga-IE": "Irish (Ireland)",
    "gd": "Scottish Gaelic",
    "gd-GB": "Scottish Gaelic (United Kingdom)",
    "gl": "Galician",
    "gl-ES": "Galician (Spain)",
    "gsw": "Swiss German",
    "gsw-CH": "Swiss German (Switzerland)",
    "gsw-FR": "Swiss German (France)",
    "gsw-LI": "Swiss German (Liechtenstein)",
    "gu": "Gujarati",
    "gu-IN": "Gujarati (India)",
    "guz": "Gusii",
    "guz-KE": "Gusii (Kenya)",
    "gv": "Manx",
    "gv-IM": "Manx (Isle of Man)",
    "ha": "Hausa",
    "ha-Latn": "Hausa (Latin)",
    "ha-Latn-GH": "Hausa (Latin, Ghana)",
    "ha-Latn-NE": "Hausa (Latin, Niger)",
    "ha-Latn-NG": "Hausa (Latin, Nigeria)",
    "haw": "Hawaiian",
    "haw-US": "Hawaiian (United States)",
    "he": "Hebrew",
    "he-IL": "Hebrew (Israel)",
    "hi": "Hindi",
    "hi-IN": "Hindi (India)",
    "hr": "Croatian",
    "hr-BA": "Croatian (Bosnia & Herzegovina)",
    "hr-HR": "Croatian (Croatia)",
    "hsb": "Upper Sorbian",
    "hsb-DE": "Upper Sorbian (Germany)",
    "hu": "Hungarian",
    "hu-HU": "Hungarian (Hungary)",
    "hy": "Armenian",
    "hy-AM": "Armenian (Armenia)",
    "id": "Indonesian",
    "id-ID": "Indonesian (Indonesia)",
    "ig": "Igbo",
    "ig-NG": "Igbo (Nigeria)",
    "ii": "Sichuan Yi",
    "ii-CN": "Sichuan Yi (China)",
    "is": "Icelandic",
    "is-IS": "Icelandic (Iceland)",
    "it": "Italian",
    "it-CH": "Italian (Switzerland)",
    "it-IT": "Italian (Italy)",
    "it-SM": "Italian (San Marino)",
    "iu": "Inuktitut",
    "iu-Cans": "Inuktitut (Unified Canadian Aboriginal Syllabics)",
    "iu-Cans-CA": "Inuktitut (Unified Canadian Aboriginal Syllabics, Canada)",
    "ja": "Japanese",
    "ja-JP": "Japanese (Japan)",
    "jgo": "Ngomba",
    "jgo-CM": "Ngomba (Cameroon)",
    "jmc": "Machame",
    "jmc-TZ": "Machame (Tanzania)",
    "ka": "Georgian",
    "ka-GE": "Georgian (Georgia)",
    "kab": "Kabyle",
    "kab-DZ": "Kabyle (Algeria)",
    "kam": "Kamba",
    "kam-KE": "Kamba (Kenya)",
    "kde": "Makonde",
    "kde-TZ": "Makonde (Tanzania)",
    "kea": "Kabuverdianu",
    "kea-CV": "Kabuverdianu (Cape Verde)",
    "khq": "Koyra Chiini",
    "khq-ML": "Koyra Chiini (Mali)",
    "ki": "Kikuyu",
    "ki-KE": "Kikuyu (Kenya)",
    "kk": "Kazakh",
    "kk-Cyrl": "Kazakh (Cyrillic)",
    "kk-Cyrl-KZ": "Kazakh (Cyrillic, Kazakhstan)",
    "kkj": "Kako",
    "kkj-CM": "Kako (Cameroon)",
    "kl": "Kalaallisut",
    "kl-GL": "Kalaallisut (Greenland)",
    "kln": "Kalenjin",
    "kln-KE": "Kalenjin (Kenya)",
    "km": "Khmer",
    "km-KH": "Khmer (Cambodia)",
    "kn": "Kannada",
    "kn-IN": "Kannada (India)",
    "ko": "Korean",
    "ko-KP": "Korean (North Korea)",
    "ko-KR": "Korean (South Korea)",
    "kok": "Konkani",
    "kok-IN": "Konkani (India)",
    "ks": "Kashmiri",
    "ks-Arab": "Kashmiri (Arabic)",
    "ks-Arab-IN": "Kashmiri (Arabic, India)",
    "ksb": "Shambala",
    "ksb-TZ": "Shambala (Tanzania)",
    "ksf": "Bafia",
    "ksf-CM": "Bafia (Cameroon)",
    "ksh": "Colognian",
    "ksh-DE": "Colognian (Germany)",
    "kw": "Cornish",
    "kw-GB": "Cornish (United Kingdom)",
    "ky": "Kyrgyz",
    "ky-Cyrl": "Kyrgyz (Cyrillic)",
    "ky-Cyrl-KG": "Kyrgyz (Cyrillic, Kyrgyzstan)",
    "lag": "Langi",
    "lag-TZ": "Langi (Tanzania)",
    "lb": "Luxembourgish",
    "lb-LU": "Luxembourgish (Luxembourg)",
    "lg": "Ganda",
    "lg-UG": "Ganda (Uganda)",
    "lkt": "Lakota",
    "lkt-US": "Lakota (United States)",
    "ln": "Lingala",
    "ln-AO": "Lingala (Angola)",
    "ln-CD": "Lingala (Congo - Kinshasa)",
    "ln-CF": "Lingala (Central African Republic)",
    "ln-CG": "Lingala (Congo - Brazzaville)",
    "lo": "Lao",
    "lo-LA": "Lao (Laos)",
    "localeIdentifier": "Description",
    "lt": "Lithuanian",
    "lt-LT": "Lithuanian (Lithuania)",
    "lu": "Luba-Katanga",
    "lu-CD": "Luba-Katanga (Congo - Kinshasa)",
    "luo": "Luo",
    "luo-KE": "Luo (Kenya)",
    "luy": "Luyia",
    "luy-KE": "Luyia (Kenya)",
    "lv": "Latvian",
    "lv-LV": "Latvian (Latvia)",
    "mas": "Masai",
    "mas-KE": "Masai (Kenya)",
    "mas-TZ": "Masai (Tanzania)",
    "mer": "Meru",
    "mer-KE": "Meru (Kenya)",
    "mfe": "Morisyen",
    "mfe-MU": "Morisyen (Mauritius)",
    "mg": "Malagasy",
    "mg-MG": "Malagasy (Madagascar)",
    "mgh": "Makhuwa-Meetto",
    "mgh-MZ": "Makhuwa-Meetto (Mozambique)",
    "mgo": "Metaʼ",
    "mgo-CM": "Metaʼ (Cameroon)",
    "mk": "Macedonian",
    "mk-MK": "Macedonian (Macedonia)",
    "ml": "Malayalam",
    "ml-IN": "Malayalam (India)",
    "mn": "Mongolian",
    "mn-Cyrl": "Mongolian (Cyrillic)",
    "mn-Cyrl-MN": "Mongolian (Cyrillic, Mongolia)",
    "mr": "Marathi",
    "mr-IN": "Marathi (India)",
    "ms": "Malay",
    "ms-Arab": "Malay (Arabic)",
    "ms-Arab-BN": "Malay (Arabic, Brunei)",
    "ms-Arab-MY": "Malay (Arabic, Malaysia)",
    "ms-Latn": "Malay (Latin)",
    "ms-Latn-BN": "Malay (Latin, Brunei)",
    "ms-Latn-MY": "Malay (Latin, Malaysia)",
    "ms-Latn-SG": "Malay (Latin, Singapore)",
    "mt": "Maltese",
    "mt-MT": "Maltese (Malta)",
    "mua": "Mundang",
    "mua-CM": "Mundang (Cameroon)",
    "my": "Burmese",
    "my-MM": "Burmese (Myanmar [Burma])",
    "naq": "Nama",
    "naq-NA": "Nama (Namibia)",
    "nb": "Norwegian Bokmål",
    "nb-NO": "Norwegian Bokmål (Norway)",
    "nb-SJ": "Norwegian Bokmål (Svalbard & Jan Mayen)",
    "nd": "North Ndebele",
    "nd-ZW": "North Ndebele (Zimbabwe)",
    "ne": "Nepali",
    "ne-IN": "Nepali (India)",
    "ne-NP": "Nepali (Nepal)",
    "nl": "Dutch",
    "nl-AW": "Dutch (Aruba)",
    "nl-BE": "Dutch (Belgium)",
    "nl-BQ": "Dutch (Caribbean Netherlands)",
    "nl-CW": "Dutch (Curaçao)",
    "nl-NL": "Dutch (Netherlands)",
    "nl-SR": "Dutch (Suriname)",
    "nl-SX": "Dutch (Sint Maarten)",
    "nmg": "Kwasio",
    "nmg-CM": "Kwasio (Cameroon)",
    "nn": "Norwegian Nynorsk",
    "nn-NO": "Norwegian Nynorsk (Norway)",
    "nnh": "Ngiemboon",
    "nnh-CM": "Ngiemboon (Cameroon)",
    "nus": "Nuer",
    "nus-SD": "Nuer (Sudan)",
    "nyn": "Nyankole",
    "nyn-UG": "Nyankole (Uganda)",
    "om": "Oromo",
    "om-ET": "Oromo (Ethiopia)",
    "om-KE": "Oromo (Kenya)",
    "or": "Oriya",
    "or-IN": "Oriya (India)",
    "os": "Ossetic",
    "os-GE": "Ossetic (Georgia)",
    "os-RU": "Ossetic (Russia)",
    "pa": "Punjabi",
    "pa-Arab": "Punjabi (Arabic)",
    "pa-Arab-PK": "Punjabi (Arabic, Pakistan)",
    "pa-Guru": "Punjabi (Gurmukhi)",
    "pa-Guru-IN": "Punjabi (Gurmukhi, India)",
    "pl": "Polish",
    "pl-PL": "Polish (Poland)",
    "ps": "Pashto",
    "ps-AF": "Pashto (Afghanistan)",
    "pt": "Portuguese",
    "pt-AO": "Portuguese (Angola)",
    "pt-BR": "Portuguese (Brazil)",
    "pt-CV": "Portuguese (Cape Verde)",
    "pt-GW": "Portuguese (Guinea-Bissau)",
    "pt-MO": "Portuguese (Macau SAR China)",
    "pt-MZ": "Portuguese (Mozambique)",
    "pt-PT": "Portuguese (Portugal)",
    "pt-ST": "Portuguese (São Tomé & Príncipe)",
    "pt-TL": "Portuguese (Timor-Leste)",
    "qu": "Quechua",
    "qu-BO": "Quechua (Bolivia)",
    "qu-EC": "Quechua (Ecuador)",
    "qu-PE": "Quechua (Peru)",
    "rm": "Romansh",
    "rm-CH": "Romansh (Switzerland)",
    "rn": "Rundi",
    "rn-BI": "Rundi (Burundi)",
    "ro": "Romanian",
    "ro-MD": "Romanian (Moldova)",
    "ro-RO": "Romanian (Romania)",
    "rof": "Rombo",
    "rof-TZ": "Rombo (Tanzania)",
    "ru": "Russian",
    "ru-BY": "Russian (Belarus)",
    "ru-KG": "Russian (Kyrgyzstan)",
    "ru-KZ": "Russian (Kazakhstan)",
    "ru-MD": "Russian (Moldova)",
    "ru-RU": "Russian (Russia)",
    "ru-UA": "Russian (Ukraine)",
    "rw": "Kinyarwanda",
    "rw-RW": "Kinyarwanda (Rwanda)",
    "rwk": "Rwa",
    "rwk-TZ": "Rwa (Tanzania)",
    "sah": "Sakha",
    "sah-RU": "Sakha (Russia)",
    "saq": "Samburu",
    "saq-KE": "Samburu (Kenya)",
    "sbp": "Sangu",
    "sbp-TZ": "Sangu (Tanzania)",
    "se": "Northern Sami",
    "se-FI": "Northern Sami (Finland)",
    "se-NO": "Northern Sami (Norway)",
    "se-SE": "Northern Sami (Sweden)",
    "seh": "Sena",
    "seh-MZ": "Sena (Mozambique)",
    "ses": "Koyraboro Senni",
    "ses-ML": "Koyraboro Senni (Mali)",
    "sg": "Sango",
    "sg-CF": "Sango (Central African Republic)",
    "shi": "Tachelhit",
    "shi-Latn": "Tachelhit (Latin)",
    "shi-Latn-MA": "Tachelhit (Latin, Morocco)",
    "shi-Tfng": "Tachelhit (Tifinagh)",
    "shi-Tfng-MA": "Tachelhit (Tifinagh, Morocco)",
    "si": "Sinhala",
    "si-LK": "Sinhala (Sri Lanka)",
    "sk": "Slovak",
    "sk-SK": "Slovak (Slovakia)",
    "sl": "Slovenian",
    "sl-SI": "Slovenian (Slovenia)",
    "smn": "Inari Sami",
    "smn-FI": "Inari Sami (Finland)",
    "sn": "Shona",
    "sn-ZW": "Shona (Zimbabwe)",
    "so": "Somali",
    "so-DJ": "Somali (Djibouti)",
    "so-ET": "Somali (Ethiopia)",
    "so-KE": "Somali (Kenya)",
    "so-SO": "Somali (Somalia)",
    "sq": "Albanian",
    "sq-AL": "Albanian (Albania)",
    "sq-MK": "Albanian (Macedonia)",
    "sq-XK": "Albanian (Kosovo)",
    "sr": "Serbian",
    "sr-Cyrl": "Serbian (Cyrillic)",
    "sr-Cyrl-BA": "Serbian (Cyrillic, Bosnia & Herzegovina)",
    "sr-Cyrl-ME": "Serbian (Cyrillic, Montenegro)",
    "sr-Cyrl-RS": "Serbian (Cyrillic, Serbia)",
    "sr-Cyrl-XK": "Serbian (Cyrillic, Kosovo)",
    "sr-Latn": "Serbian (Latin)",
    "sr-Latn-BA": "Serbian (Latin, Bosnia & Herzegovina)",
    "sr-Latn-ME": "Serbian (Latin, Montenegro)",
    "sr-Latn-RS": "Serbian (Latin, Serbia)",
    "sr-Latn-XK": "Serbian (Latin, Kosovo)",
    "sv": "Swedish",
    "sv-AX": "Swedish (Åland Islands)",
    "sv-FI": "Swedish (Finland)",
    "sv-SE": "Swedish (Sweden)",
    "sw": "Swahili",
    "sw-CD": "Swahili (Congo - Kinshasa)",
    "sw-KE": "Swahili (Kenya)",
    "sw-TZ": "Swahili (Tanzania)",
    "sw-UG": "Swahili (Uganda)",
    "ta": "Tamil",
    "ta-IN": "Tamil (India)",
    "ta-LK": "Tamil (Sri Lanka)",
    "ta-MY": "Tamil (Malaysia)",
    "ta-SG": "Tamil (Singapore)",
    "te": "Telugu",
    "te-IN": "Telugu (India)",
    "teo": "Teso",
    "teo-KE": "Teso (Kenya)",
    "teo-UG": "Teso (Uganda)",
    "tg": "Tajik",
    "tg-Cyrl": "Tajik (Cyrillic)",
    "tg-Cyrl-TJ": "Tajik (Cyrillic, Tajikistan)",
    "th": "Thai",
    "th-TH": "Thai (Thailand)",
    "ti": "Tigrinya",
    "ti-ER": "Tigrinya (Eritrea)",
    "ti-ET": "Tigrinya (Ethiopia)",
    "tk": "Turkmen",
    "tk-Latn": "Turkmen (Latin)",
    "tk-Latn-TM": "Turkmen (Latin, Turkmenistan)",
    "to": "Tongan",
    "to-TO": "Tongan (Tonga)",
    "tr": "Turkish",
    "tr-CY": "Turkish (Cyprus)",
    "tr-TR": "Turkish (Turkey)",
    "twq": "Tasawaq",
    "twq-NE": "Tasawaq (Niger)",
    "tzm": "Central Atlas Tamazight",
    "tzm-Latn": "Central Atlas Tamazight (Latin)",
    "tzm-Latn-MA": "Central Atlas Tamazight (Latin, Morocco)",
    "ug": "Uyghur",
    "ug-Arab": "Uyghur (Arabic)",
    "ug-Arab-CN": "Uyghur (Arabic, China)",
    "uk": "Ukrainian",
    "uk-UA": "Ukrainian (Ukraine)",
    "ur": "Urdu",
    "ur-IN": "Urdu (India)",
    "ur-PK": "Urdu (Pakistan)",
    "uz": "Uzbek",
    "uz-Arab": "Uzbek (Arabic)",
    "uz-Arab-AF": "Uzbek (Arabic, Afghanistan)",
    "uz-Cyrl": "Uzbek (Cyrillic)",
    "uz-Cyrl-UZ": "Uzbek (Cyrillic, Uzbekistan)",
    "uz-Latn": "Uzbek (Latin)",
    "uz-Latn-UZ": "Uzbek (Latin, Uzbekistan)",
    "vai": "Vai",
    "vai-Latn": "Vai (Latin)",
    "vai-Latn-LR": "Vai (Latin, Liberia)",
    "vai-Vaii": "Vai (Vai)",
    "vai-Vaii-LR": "Vai (Vai, Liberia)",
    "vi": "Vietnamese",
    "vi-VN": "Vietnamese (Vietnam)",
    "vun": "Vunjo",
    "vun-TZ": "Vunjo (Tanzania)",
    "wae": "Walser",
    "wae-CH": "Walser (Switzerland)",
    "xog": "Soga",
    "xog-UG": "Soga (Uganda)",
    "yav": "Yangben",
    "yav-CM": "Yangben (Cameroon)",
    "yi": "Yiddish",
    "yi-001": "Yiddish (World)",
    "yo": "Yoruba",
    "yo-BJ": "Yoruba (Benin)",
    "yo-NG": "Yoruba (Nigeria)",
    "zgh": "Standard Moroccan Tamazight",
    "zgh-MA": "Standard Moroccan Tamazight (Morocco)",
    "zh": "Chinese",
    "zh-Hans": "Chinese (Simplified)",
    "zh-Hans-CN": "Chinese (Simplified, China)",
    "zh-Hans-HK": "Chinese (Simplified, Hong Kong SAR China)",
    "zh-Hans-MO": "Chinese (Simplified, Macau SAR China)",
    "zh-Hans-SG": "Chinese (Simplified, Singapore)",
    "zh-Hant": "Chinese (Traditional)",
    "zh-Hant-HK": "Chinese (Traditional, Hong Kong SAR China)",
    "zh-Hant-MO": "Chinese (Traditional, Macau SAR China)",
    "zh-Hant-TW": "Chinese (Traditional, Taiwan)",
    "zu": "Zulu",
    "zu-ZA": "Zulu (South Africa)",
}



from __future__ import annotations

import socket
from dataclasses import dataclass


def get_lan_ip():
    return socket.gethostbyname(socket.gethostname())


@dataclass(frozen=True, slots=True)
class Address:
    host: str
    port: int
    secure: bool = False
    hash: str | None = None

    def with_hash(self, hash: str | None):
        return Address(
            host=self.host,
            port=self.port,
            secure=self.secure,
            hash=hash,
        )

    @classmethod
    def default(cls) -> Address:
        return cls(host=get_lan_ip(), port=26423)

    def to_url(self) -> str:
        return f"{'https' if self.secure else 'http'}://" f"{self.host or 'localhost'}:{self.port}"



from .extension import (
    DASHBOARD_APP_INSTALL_ENDPOINT,
    DASHBOARD_APP_INSTALL_PACKET,
    DASHBOARD_APP_INSTALL_PERMISSION_ID,
    DASHBOARD_APP_UPDATE_PACKET,
    DASHBOARD_DRAG_DROP_PERMISSION_ID,
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    DASHBOARD_SET_PERMISSION_ID,
    DASHBOARD_SPEECH_RECOGNITION_PERMISSION_ID,
    DASHBOARD_WEBVIEW_PERMISSION_ID,
)
from .packets import PermissionRequestPacket

__all__ = [
    "DASHBOARD_APP_INSTALL_ENDPOINT",
    "DASHBOARD_APP_INSTALL_PACKET",
    "DASHBOARD_APP_INSTALL_PERMISSION_ID",
    "DASHBOARD_APP_UPDATE_PACKET",
    "DASHBOARD_OPEN_APP_PERMISSION_ID",
    "DASHBOARD_SET_PERMISSION_ID",
    "DASHBOARD_SPEECH_RECOGNITION_PERMISSION_ID",
    "DASHBOARD_WEBVIEW_PERMISSION_ID",
    "DASHBOARD_DRAG_DROP_PERMISSION_ID",
    "PermissionRequestPacket",
]



from __future__ import annotations

import io
import json
import struct
from collections.abc import Callable
from typing import Any


class Flags:
    def __init__(self, value: int = 0, length: int = 32) -> None:
        self.value = value & ((1 << length) - 1)
        self.length = length

    def has(self, position: int) -> bool:
        return bool(self.value & (1 << position))

    def get(self, position: int) -> bool:
        return bool(self.value & (1 << position))

    def if_set[T](self, position: int, callback: Callable[[], T]) -> T | None:
        if self.has(position):
            return callback()
        return None

    def set(self, position: int, value: bool = True) -> Flags:
        if value:
            self.value |= 1 << position
        else:
            self.value &= ~(1 << position)
        return self

    def unset(self, position: int) -> Flags:
        self.value &= ~(1 << position)
        return self

    def write(self, writer: ByteWriter) -> ByteWriter:
        bits = self.value.to_bytes((self.length + 7) // 8, "big")
        writer.write(bits)
        return writer

    @classmethod
    def read(cls, reader: ByteReader, length: int) -> Flags:
        bits = int.from_bytes(reader.read((length + 7) // 8), "big")
        return Flags(bits, length)

    def __or__(self, other: Flags) -> Flags:
        return Flags(self.value | other.value, max(self.length, other.length))

    def __and__(self, other: Flags) -> Flags:
        return Flags(self.value & other.value, max(self.length, other.length))

    def __xor__(self, other: Flags) -> Flags:
        return Flags(self.value ^ other.value, max(self.length, other.length))

    def __invert__(self) -> Flags:
        return Flags(~self.value, self.length)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Flags):
            return False
        return self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)


type ByteArray = bytearray | bytes


class ByteWriter:
    def __init__(self, init: ByteArray | None = None) -> None:
        self.stream = io.BytesIO(init or b"")
        self.finished = False

    def write(self, data: ByteArray) -> ByteWriter:
        if self.finished:
            raise ValueError("Writer already finished")
        self.stream.write(data)
        return self

    def write_boolean(self, value: bool) -> ByteWriter:
        self.write(value.to_bytes(1, "big"))
        return self

    def write_int8(self, value: int) -> ByteWriter:
        if not (-128 <= value < 128):
            raise ValueError("Value must be in range -128 to 127")
        self.write(value.to_bytes(1, "big", signed=True))
        return self

    def write_uint8(self, value: int) -> ByteWriter:
        if not (0 <= value < 256):
            raise ValueError("Value must be in range 0-255")
        self.write(value.to_bytes(1, "big"))
        return self

    def write_int16(self, value: int) -> ByteWriter:
        if not (-32768 <= value < 32768):
            raise ValueError("Value must be in range -32768 to 32767")
        self.write(value.to_bytes(2, "big", signed=True))
        return self

    def write_uint16(self, value: int) -> ByteWriter:
        if not (0 <= value < 65536):
            raise ValueError("Value must be in range 0-65535")
        self.write(value.to_bytes(2, "big"))
        return self

    def write_int32(self, value: int) -> ByteWriter:
        if not (-2147483648 <= value < 2147483648):
            raise ValueError("Value must be in range -2147483648 to 2147483647")
        self.write(value.to_bytes(4, "big", signed=True))
        return self

    def write_uint32(self, value: int) -> ByteWriter:
        if not (0 <= value < 4294967296):
            raise ValueError("Value must be in range 0-4294967295")
        self.write(value.to_bytes(4, "big"))
        return self

    def write_int64(self, value: int) -> ByteWriter:
        if not (-9223372036854775808 <= value < 9223372036854775808):
            raise ValueError("Value must be in range -9223372036854775808 to 9223372036854775807")
        self.write(value.to_bytes(8, "big", signed=True))
        return self

    def write_uint64(self, value: int) -> ByteWriter:
        if not (0 <= value < 18446744073709551616):
            raise ValueError("Value must be in range 0-18446744073709551615")
        self.write(value.to_bytes(8, "big"))
        return self

    def write_uleb128(self, value: int) -> ByteWriter:
        if value < 0:
            raise ValueError("Value must be non-negative")
        while value > 127:
            self.write(bytes([(value & 0x7F) | 0x80]))
            value >>= 7
        self.write(bytes([value & 0x7F]))
        return self

    def write_float16(self, value: float) -> ByteWriter:
        if not (-65504.0 <= value <= 65504.0):
            raise ValueError("Value must be in range -65504.0 to 65504.0")
        self.write(struct.pack(">e", value))
        return self

    def write_float32(self, value: float) -> ByteWriter:
        self.write(struct.pack(">f", value))
        return self

    def write_float64(self, value: float) -> ByteWriter:
        self.write(struct.pack(">d", value))
        return self

    def write_uint8_array(self, values: ByteArray) -> ByteWriter:
        self.write_uleb128(len(values))
        self.write(values)
        return self

    def write_string(self, value: str) -> ByteWriter:
        encoded = value.encode("utf-8")
        self.write_uleb128(len(encoded))
        self.write(encoded)
        return self

    def write_json(self, value: Any) -> ByteWriter:
        self.write_string(json.dumps(value))
        return self

    def write_flags(self, flags: Flags) -> ByteWriter:
        flags.write(self)
        return self

    def finish(self) -> bytes:
        if self.finished:
            raise ValueError("Writer already finished")
        self.finished = True
        return self.stream.getvalue()


class ByteReader:
    def __init__(self, buffer: ByteArray) -> None:
        self.stream = io.BytesIO(buffer)
        self.is_reading = False
        self.is_finished = False

    def __enter__(self) -> ByteReader:
        if self.is_reading:
            raise ValueError("Reader already reading")
        if self.is_finished:
            raise ValueError("Reader already finished")
        self.is_reading = True
        return self

    def finish(self) -> None:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        if self.stream.read(1):
            raise ValueError("Reader not fully consumed")
        self.is_reading = False
        self.is_finished = True

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        self.finish()

    def read(self, size: int) -> bytes:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        if size < 0:
            raise ValueError("Size must be positive")
        return self.stream.read(size)

    def read_boolean(self) -> bool:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return bool(int.from_bytes(self.read(1), "big"))

    def read_int8(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return int.from_bytes(self.read(1), "big", signed=True)

    def read_uint8(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return int.from_bytes(self.read(1), "big")

    def read_int16(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return int.from_bytes(self.read(2), "big", signed=True)

    def read_uint16(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return int.from_bytes(self.read(2), "big")

    def read_int32(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return int.from_bytes(self.read(4), "big", signed=True)

    def read_uint32(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return int.from_bytes(self.read(4), "big")

    def read_int64(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return int.from_bytes(self.read(8), "big", signed=True)

    def read_uint64(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return int.from_bytes(self.read(8), "big")

    def read_uleb128(self) -> int:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        value = 0
        shift = 0
        while True:
            byte = self.read(1)[0]
            value |= (byte & 0x7F) << shift
            if not (byte & 0x80):
                break
            shift += 7
        return value

    def read_float16(self) -> float:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return struct.unpack(">e", self.read(2))[0]

    def read_float32(self) -> float:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return struct.unpack(">f", self.read(4))[0]

    def read_float64(self) -> float:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return struct.unpack(">d", self.read(8))[0]

    def read_uint8_array(self) -> bytes:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        length = self.read_uleb128()
        return self.read(length)

    def read_string(self) -> str:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        length = self.read_uleb128()
        return self.read(length).decode("utf-8")

    def read_json[T](self) -> T:  # type: ignore
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return json.loads(self.read_string())

    def read_flags(self, length: int) -> Flags:
        if not self.is_reading:
            raise ValueError("Reader not reading")
        return Flags.read(self, length)



from __future__ import annotations

import abc


class Model[D](abc.ABC):
    @abc.abstractmethod
    def to_json(self) -> D: ...

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json: D) -> Model[D]: ...



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

from omu.api import Extension, ExtensionType
from omu.api.endpoint import EndpointType
from omu.api.table import TableType
from omu.api.table.table import TablePermissions
from omu.network.packet import PacketType

if TYPE_CHECKING:
    from omu.omu import Omu

from .plugin import PluginPackageInfo

PLUGIN_EXTENSION_TYPE = ExtensionType("plugin", lambda client: PluginExtension(client))


class PluginExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return PLUGIN_EXTENSION_TYPE

    def __init__(self, omu: Omu):
        self.omu = omu
        self.plugins: dict[str, str | None] = {}

        self.omu.network.register_packet(
            PLUGIN_REQUIRE_PACKET,
        )
        self.omu.network.add_task(self.on_task)

    async def on_task(self):
        await self.omu.send(PLUGIN_REQUIRE_PACKET, self.plugins)

    def require(self, plugins: dict[str, str | None]):
        if self.omu.running:
            raise RuntimeError("Cannot require plugins after client has started")
        self.plugins.update(plugins)


PLUGIN_REQUIRE_PACKET = PacketType[dict[str, str | None]].create_json(
    PLUGIN_EXTENSION_TYPE,
    "require",
)
PLUGIN_READ_PACKAGE_PERMISSION_ID = PLUGIN_EXTENSION_TYPE.join("package", "read")
PLUGIN_MANAGE_PACKAGE_PERMISSION_ID = PLUGIN_EXTENSION_TYPE.join("package", "manage")
PLUGIN_ALLOWED_PACKAGE_TABLE = TableType.create_model(
    PLUGIN_EXTENSION_TYPE,
    "allowed_package",
    PluginPackageInfo,
    permissions=TablePermissions(
        read=PLUGIN_READ_PACKAGE_PERMISSION_ID,
    ),
)


class ReloadOptions(TypedDict):
    packages: list[str] | None


class ReloadResult(TypedDict):
    packages: dict[str, str]


PLUGIN_RELOAD_ENDPOINT_TYPE = EndpointType[ReloadOptions, ReloadResult].create_json(
    PLUGIN_EXTENSION_TYPE,
    name="reload",
    permission_id=PLUGIN_MANAGE_PACKAGE_PERMISSION_ID,
)



from __future__ import annotations

from enum import Enum
from typing import Final, NotRequired, TypedDict

from omu.helper import map_optional
from omu.identifier import Identifier
from omu.interface import Keyable
from omu.localization import Locale, LocalizedText
from omu.model import Model


class AppMetadata(TypedDict):
    locale: Locale
    name: NotRequired[LocalizedText | None]
    icon: NotRequired[LocalizedText | None]
    description: NotRequired[LocalizedText | None]
    image: NotRequired[LocalizedText | None]
    site: NotRequired[LocalizedText | None]
    repository: NotRequired[LocalizedText | None]
    authors: NotRequired[LocalizedText | None]
    license: NotRequired[LocalizedText | None]
    tags: NotRequired[list[str] | None]


class AppType(Enum):
    APP = "app"
    SERVICE = "service"
    REMOTE = "remote"
    PLUGIN = "plugin"
    DASHBOARD = "dashboard"


class DependencySpecifier(TypedDict):
    version: str
    index: NotRequired[str] | None


class AppJson(TypedDict):
    id: str
    type: str
    parent_id: NotRequired[str | None]
    version: NotRequired[str | None]
    url: NotRequired[str | None]
    metadata: NotRequired[AppMetadata | None]
    dependencies: NotRequired[dict[str, DependencySpecifier | str] | None]


class App(Keyable, Model[AppJson]):
    def __init__(
        self,
        id: Identifier | str,
        *,
        version: str | None = None,
        url: str | None = None,
        type: AppType = AppType.APP,
        parent_id: Identifier | None = None,
        metadata: AppMetadata | None = None,
        dependencies: dict[str, DependencySpecifier | str] | None = None,
    ) -> None:
        if isinstance(id, str):
            id = Identifier.from_key(id)
        self.parent_id = parent_id
        self.id: Final[Identifier] = id
        self.version = version
        self.url = url
        self.type = type
        self.metadata = metadata
        self.dependencies = dependencies

    @classmethod
    def from_json(cls, json: AppJson) -> App:
        id = Identifier.from_key(json["id"])
        return cls(
            id,
            parent_id=map_optional(json.get("parent_id"), Identifier.from_key),
            version=json.get("version"),
            url=json.get("url"),
            type=AppType(json.get("type", None) or "app"),
            metadata=json.get("metadata"),
            dependencies=json.get("dependencies"),
        )

    def to_json(self) -> AppJson:
        return AppJson(
            id=self.key(),
            parent_id=self.parent_id.key() if self.parent_id else None,
            version=self.version,
            url=self.url,
            type=self.type.value,
            metadata=self.metadata,
            dependencies=self.dependencies,
        )

    def key(self) -> str:
        return self.id.key()

    def __hash__(self) -> int:
        return hash(self.key())

    def __repr__(self) -> str:
        return f"App({self.key()})"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Literal, NotRequired, TypedDict

from omu.api.endpoint.endpoint import EndpointType
from omu.api.extension import Extension, ExtensionType
from omu.api.table.table import TablePermissions, TableType
from omu.app import App, AppJson
from omu.helper import Coro
from omu.identifier import Identifier
from omu.localization.locale import Locale
from omu.localization.localization import LocalizedText
from omu.network.packet import PacketType
from omu.omu import Omu
from omu.serializer import Serializer

SESSION_EXTENSION_TYPE = ExtensionType("sessions", lambda client: SessionExtension(client))
SESSIONS_READ_PERMISSION_ID = SESSION_EXTENSION_TYPE / "read"

SESSION_TABLE_TYPE = TableType.create_model(
    SESSION_EXTENSION_TYPE,
    "sessions",
    App,
    permissions=TablePermissions(
        read=SESSIONS_READ_PERMISSION_ID,
    ),
)

SESSION_OBSERVE_PACKET_TYPE = PacketType[list[Identifier]].create_json(
    SESSION_EXTENSION_TYPE,
    "observe",
    serializer=Serializer.model(Identifier).to_array(),
)
SESSION_CONNECTED_PACKET_TYPE = PacketType[App].create_json(
    SESSION_EXTENSION_TYPE,
    "session_connected",
    serializer=Serializer.model(App),
)
SESSION_DISCONNECTED_PACKET_TYPE = PacketType[App].create_json(
    SESSION_EXTENSION_TYPE,
    "session_disconnected",
    serializer=Serializer.model(App),
)


class RemoteAppMetadata(TypedDict):
    locale: Locale
    name: NotRequired[LocalizedText | None]
    icon: NotRequired[LocalizedText | None]
    description: NotRequired[LocalizedText | None]


class RemoteAppRequestPayload(TypedDict):
    app: AppJson
    permissions: list[str]


class RequestRemoteAppResponseOk(TypedDict):
    type: Literal["success"]
    token: str
    lan_ip: str


class RequestRemoteAppResponseError(TypedDict):
    type: Literal["error"]
    message: str


type RequestRemoteAppResponse = RequestRemoteAppResponseOk | RequestRemoteAppResponseError

REMOTE_APP_REQUEST_PERMISSION_ID = SESSION_EXTENSION_TYPE / "remote_app" / "request"
REMOTE_APP_REQUEST_ENDPOINT_TYPE = EndpointType[RemoteAppRequestPayload, RequestRemoteAppResponse].create_json(
    SESSION_EXTENSION_TYPE,
    "remote_app_request",
    permission_id=REMOTE_APP_REQUEST_PERMISSION_ID,
)
SESSION_REQUIRE_PACKET_TYPE = PacketType[list[Identifier]].create_json(
    SESSION_EXTENSION_TYPE,
    "require",
    serializer=Serializer.model(Identifier).to_array(),
)


class GenerateTokenPayload(TypedDict):
    app: AppJson
    permissions: list[str]


class GenerateTokenResponseOk(TypedDict):
    type: Literal["success"]
    token: str


class GenerateTokenResponseError(TypedDict):
    type: Literal["error"]
    message: str


type GenerateTokenResponse = GenerateTokenResponseOk | GenerateTokenResponseError

GENERATE_TOKEN_PERMISSION_ID = SESSION_EXTENSION_TYPE / "generate_token"
GENERATE_TOKEN_ENDPOINT_TYPE = EndpointType[GenerateTokenPayload, GenerateTokenResponse].create_json(
    SESSION_EXTENSION_TYPE,
    "generate_token",
    permission_id=GENERATE_TOKEN_PERMISSION_ID,
)


class SessionExtension(Extension):
    @property
    def type(self):
        return SESSION_EXTENSION_TYPE

    def __init__(self, omu: Omu):
        self.omu = omu
        omu.network.register_packet(
            SESSION_REQUIRE_PACKET_TYPE,
            SESSION_OBSERVE_PACKET_TYPE,
            SESSION_CONNECTED_PACKET_TYPE,
            SESSION_DISCONNECTED_PACKET_TYPE,
        )
        omu.network.add_packet_handler(SESSION_CONNECTED_PACKET_TYPE, self.handle_session_connect)
        omu.network.add_packet_handler(SESSION_DISCONNECTED_PACKET_TYPE, self.handle_session_disconnect)
        self.session_observers: dict[Identifier, SessionObserver] = {}
        self.required_apps: set[Identifier] = set()
        omu.network.add_task(self.on_task)
        omu.on_ready(self.on_ready)

    async def on_ready(self) -> None:
        if self.session_observers:
            await self.omu.send(SESSION_OBSERVE_PACKET_TYPE, [*self.session_observers])

    async def on_task(self) -> None:
        if self.required_apps:
            await self.omu.send(SESSION_REQUIRE_PACKET_TYPE, [*self.required_apps])

    def require(self, *app_ids: Identifier) -> None:
        if self.omu.running:
            raise RuntimeError("Cannot require apps after the client has started")
        self.required_apps.update(app_ids)

    async def handle_session_connect(self, app: App) -> None:
        observer = self.session_observers.get(app.id)
        if observer is None:
            return
        for callback in observer.on_connect_callbacks:
            await callback(app)

    async def handle_session_disconnect(self, app: App) -> None:
        observer = self.session_observers.get(app.id)
        if observer is None:
            return
        for callback in observer.on_disconnect_callbacks:
            await callback(app)

    def observe_session(
        self,
        app_id: Identifier,
        on_connect: Coro[[App], None] | None = None,
        on_disconnect: Coro[[App], None] | None = None,
    ) -> SessionObserver:
        if self.omu.running:
            raise RuntimeError("Cannot require apps after the client has started")
        observer = self.session_observers.get(app_id) or SessionObserver()
        if on_connect:
            observer.on_connect(on_connect)
        if on_disconnect:
            observer.on_disconnect(on_disconnect)
        self.session_observers[app_id] = observer
        return observer


@dataclass(frozen=True, slots=True)
class SessionObserver:
    on_connect_callbacks: list[Coro[[App], None]] = field(default_factory=list)
    on_disconnect_callbacks: list[Coro[[App], None]] = field(default_factory=list)

    def on_connect(self, coro: Coro[[App], None]) -> Coro[[App], None]:
        self.on_connect_callbacks.append(coro)
        return coro

    def on_disconnect(self, coro: Coro[[App], None]) -> Coro[[App], None]:
        self.on_disconnect_callbacks.append(coro)
        return coro



from __future__ import annotations

import abc
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, NotRequired, TypedDict

from omu.event_emitter import Unlisten
from omu.helper import Coro, map_optional
from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


class RegistryPermissionsJSON(TypedDict):
    all: NotRequired[str | None]
    read: NotRequired[str | None]
    write: NotRequired[str | None]


@dataclass(frozen=True, slots=True)
class RegistryPermissions:
    all: Identifier | None = None
    read: Identifier | None = None
    write: Identifier | None = None

    @classmethod
    def from_json(cls, data: RegistryPermissionsJSON) -> RegistryPermissions:
        return RegistryPermissions(
            map_optional(data.get("all"), Identifier.from_key),
            map_optional(data.get("read"), Identifier.from_key),
            map_optional(data.get("write"), Identifier.from_key),
        )

    def to_json(self) -> RegistryPermissionsJSON:
        return RegistryPermissionsJSON(
            all=map_optional(self.all, Identifier.key),
            read=map_optional(self.read, Identifier.key),
            write=map_optional(self.write, Identifier.key),
        )


@dataclass(frozen=True, slots=True)
class RegistryType[T]:
    id: Identifier
    default_value: T
    serializer: Serializable[T, bytes]
    permissions: RegistryPermissions = RegistryPermissions()

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
        default_value: T,
        permissions: RegistryPermissions | None = None,
    ) -> RegistryType[T]:
        return cls(
            identifier / name,
            default_value,
            Serializer.json(),
            permissions or RegistryPermissions(),
        )

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        default_value: T,
        serializer: Serializable[T, bytes],
        permissions: RegistryPermissions | None = None,
    ) -> RegistryType[T]:
        return cls(
            identifier / name,
            default_value,
            serializer,
            permissions or RegistryPermissions(),
        )


class Registry[T](abc.ABC):
    @property
    @abc.abstractmethod
    def value(self) -> T: ...

    @abc.abstractmethod
    async def get(self) -> T: ...

    @abc.abstractmethod
    async def set(self, value: T) -> None: ...

    @abc.abstractmethod
    async def update(self, handler: Coro[[T], T] | Callable[[T], T]) -> T: ...

    @abc.abstractmethod
    async def modify(self, handler: Coro[[T], Any] | Callable[[T], Any]) -> T: ...

    @abc.abstractmethod
    def listen(self, handler: Coro[[T], None]) -> Unlisten: ...



from omu.api import Extension, ExtensionType
from omu.api.registry import RegistryType
from omu.api.registry.packets import RegistryPermissions
from omu.localization import Locale, LocalizedText
from omu.omu import Omu

I18N_EXTENSION_TYPE = ExtensionType("i18n", lambda client: I18nExtension(client))
I18N_SET_LOCALES_PERMISSION_ID = I18N_EXTENSION_TYPE / "locales" / "set"
I18N_GET_LOCALES_PERMISSION_ID = I18N_EXTENSION_TYPE / "locales" / "get"
I18N_LOCALES_REGISTRY_TYPE = RegistryType[list[Locale]].create_json(
    I18N_EXTENSION_TYPE,
    name="locales",
    default_value=[],
    permissions=RegistryPermissions(
        read=I18N_GET_LOCALES_PERMISSION_ID,
        write=I18N_SET_LOCALES_PERMISSION_ID,
    ),
)


class I18nExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return I18N_EXTENSION_TYPE

    def __init__(self, omu: Omu):
        omu.permissions.require(I18N_GET_LOCALES_PERMISSION_ID)
        self.locales_registry = omu.registries.get(I18N_LOCALES_REGISTRY_TYPE)
        self.locales: list[Locale] = []
        self.default_locales: list[Locale] = []

    def get_locales(self) -> list[Locale]:
        if self.locales:
            return self.locales
        if not self.default_locales:
            raise ValueError("Default locales are not set")
        return self.default_locales

    def translate(self, localized_text: LocalizedText) -> str:
        locales = self.get_locales()
        if not locales:
            raise RuntimeError("Locales not loaded")
        if isinstance(localized_text, str):
            return localized_text
        translation = self.select_best_translation(locales, localized_text)
        if not translation:
            raise ValueError(f"Missing translation for {locales} in {localized_text}")
        return translation

    def select_best_translation(
        self,
        locales: list[Locale],
        localized_text: LocalizedText,
    ) -> str | None:
        if isinstance(localized_text, str):
            return localized_text
        if not localized_text:
            return None
        translations = localized_text
        for locale in locales:
            translation = translations.get(locale)
            if translation:
                return translation
        translation = next(iter(translations.values()))
        return translation



from omu.identifier import Identifier
from omu.network.packet.packet_types import DisconnectType


class OmuError(Exception):
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class DisconnectReason(OmuError):
    def __init__(self, type: DisconnectType, message: str | None = None):
        super().__init__(message or type.value)
        self.type = type
        self.message = message


class NetworkError(OmuError):
    pass


class AnotherConnection(NetworkError, DisconnectReason):
    def __init__(self, *args, **kwargs):
        super().__init__(DisconnectType.ANOTHER_CONNECTION, *args, **kwargs)


class PermissionDenied(NetworkError, DisconnectReason):
    def __init__(self, *args, **kwargs):
        super().__init__(DisconnectType.PERMISSION_DENIED, *args, **kwargs)


class InvalidToken(NetworkError, DisconnectReason):
    def __init__(self, *args, **kwargs):
        super().__init__(DisconnectType.INVALID_TOKEN, *args, **kwargs)


class InvalidOrigin(NetworkError, DisconnectReason):
    def __init__(self, *args, **kwargs):
        super().__init__(DisconnectType.INVALID_ORIGIN, *args, **kwargs)


class InvalidVersion(NetworkError, DisconnectReason):
    def __init__(self, *args, **kwargs):
        super().__init__(DisconnectType.INVALID_VERSION, *args, **kwargs)


class InvalidPacket(NetworkError, DisconnectReason):
    def __init__(self, *args, **kwargs):
        super().__init__(DisconnectType.INVALID_PACKET, *args, **kwargs)


class InvalidPacketType(InvalidPacket):
    def __init__(self, packet_type: Identifier):
        message = f"Invalid packet type {packet_type}"
        super().__init__(DisconnectType.INVALID_PACKET_TYPE, message)
        self.packet_type = packet_type


class InvalidPacketData(InvalidPacket):
    def __init__(self, packet_type: Identifier, message: str):
        message = f"Invalid packet data for {packet_type}: {message}"
        super().__init__(DisconnectType.INVALID_PACKET_DATA, message)
        self.packet_type = packet_type



from __future__ import annotations

import io
import random
import time
from asyncio import Future
from collections.abc import Callable
from dataclasses import dataclass
from typing import Literal, NotRequired, TypedDict

from loguru import logger
from yarl import URL

from omu.api.extension import Extension, ExtensionType
from omu.bytebuffer import ByteReader, ByteWriter
from omu.network.packet.packet import PacketType
from omu.omu import Omu

HTTP_EXTENSION_TYPE = ExtensionType("http", lambda client: HttpExtension(client))

HTTP_REQUEST_PERMISSION_ID = HTTP_EXTENSION_TYPE / "request"


class RequestHandle(TypedDict):
    id: str


type RequestRedirect = Literal["error", "follow", "manual"]


class HttpRequest(RequestHandle):
    header: dict[str, str]
    method: str
    redirect: RequestRedirect
    url: str


HTTP_REQUEST_CREATE = PacketType[HttpRequest].create_json(
    HTTP_EXTENSION_TYPE,
    name="request_create",
)


@dataclass
class DataChunk[T]:
    meta: T
    data: bytes

    @staticmethod
    def serialize(data: DataChunk[T]):
        writer = ByteWriter()
        writer.write_json(data.meta)
        writer.write_uint8_array(data.data)
        return writer.finish()

    @staticmethod
    def deserialize(data: bytes) -> DataChunk[T]:
        with ByteReader(data) as reader:
            meta: T = reader.read_json()
            body = reader.read_uint8_array()
        return DataChunk(meta, body)


HTTP_REQUEST_SEND = PacketType[DataChunk[RequestHandle]].create_serialized(
    HTTP_EXTENSION_TYPE,
    name="request_send",
    serializer=DataChunk,
)

type HttpRequestClose = RequestHandle

HTTP_REQUEST_CLOSE = PacketType[HttpRequestClose].create_json(
    HTTP_EXTENSION_TYPE,
    name="request_close",
)


class HttpResponse(RequestHandle):
    header: dict[str, str]
    status: int
    statusText: str | None
    url: str
    history: list[HttpResponse]
    redirected: bool


HTTP_RESPONSE_CREATE = PacketType[HttpResponse].create_json(
    HTTP_EXTENSION_TYPE,
    name="response_create",
)

HTTP_RESPONSE_CHUNK = PacketType[DataChunk[RequestHandle]].create_serialized(
    HTTP_EXTENSION_TYPE,
    name="response_chunk",
    serializer=DataChunk,
)

type HttpResponseClose = RequestHandle

HTTP_RESPONSE_CLOSE = PacketType[HttpResponseClose].create_json(
    HTTP_EXTENSION_TYPE,
    name="response_close",
)


class HandleStateReceiving(TypedDict):
    type: Literal["receiving"]
    receive: Callable[[bytes]]
    close: Callable[[]]


class HandleStateCreated(TypedDict):
    type: Literal["created"]
    setResponse: Callable[[HttpResponse]]


type HandleState = HandleStateCreated | HandleStateReceiving

WEBSOCKET_CREATE = PacketType[HttpRequest].create_json(
    HTTP_EXTENSION_TYPE,
    name="ws_create",
)


class WebSocketOpen(RequestHandle):
    url: str
    protocol: str | None


WEBSOCKET_OPEN = PacketType[WebSocketOpen].create_json(
    HTTP_EXTENSION_TYPE,
    name="ws_open",
)


class WSDataMeta(TypedDict):
    id: str
    type: int


WEBSOCKET_DATA = PacketType[DataChunk[WSDataMeta]].create_serialized(
    HTTP_EXTENSION_TYPE,
    name="ws_data",
    serializer=DataChunk,
)


class WebSocketClose(RequestHandle):
    code: NotRequired[int]
    reason: NotRequired[str] | None


WEBSOCKET_CLOSE = PacketType[WebSocketClose].create_json(
    HTTP_EXTENSION_TYPE,
    name="ws_close",
)


class WebSocketError(RequestHandle):
    type: Literal["ConnectionRefused"]
    reason: NotRequired[str] | None


WEBSOCKET_ERROR = PacketType[WebSocketError].create_json(
    HTTP_EXTENSION_TYPE,
    name="ws_error",
)


class HttpExtension(Extension):
    @property
    def type(self):
        return HTTP_EXTENSION_TYPE

    handles: dict[str, HandleState] = {}

    def __init__(self, omu: Omu):
        self.omu = omu
        omu.network.register_packet(
            HTTP_REQUEST_CREATE,
            HTTP_REQUEST_SEND,
            HTTP_REQUEST_CLOSE,
            HTTP_RESPONSE_CREATE,
            HTTP_RESPONSE_CHUNK,
            HTTP_RESPONSE_CLOSE,
        )
        omu.network.add_packet_handler(HTTP_RESPONSE_CREATE, self.handle_response_create)
        omu.network.add_packet_handler(HTTP_RESPONSE_CHUNK, self.handle_response_chunk)
        omu.network.add_packet_handler(HTTP_RESPONSE_CLOSE, self.handle_response_close)

    async def handle_response_create(self, packet: HttpResponse):
        handle = self.handles.get(packet["id"])
        if handle is None:
            logger.warning("Received response for unknown request", packet["id"])
            return
        if handle["type"] != "created":
            logger.warning("Received response for already handled request", packet["id"])
            return
        handle["setResponse"](packet)

    async def handle_response_chunk(self, packet: DataChunk[RequestHandle]):
        handle = self.handles.get(packet.meta["id"])
        if handle is None:
            logger.warning("Received response for unknown request", packet.meta["id"])
            return
        if handle["type"] != "receiving":
            logger.warning("Received chunk for non-receiving request", packet.meta["id"])
            return
        handle["receive"](packet.data)

    async def handle_response_close(self, packet: HttpResponseClose):
        handle = self.handles.get(packet["id"])
        if handle is None:
            logger.warning("Received response for unknown request", packet["id"])
            return
        if handle["type"] != "receiving":
            logger.warning("Received chunk for non-receiving request", packet["id"])
            return
        handle["close"]()

    def generate_id(self) -> str:
        rnd = int(time.time_ns() - random.random() * 1e12)
        id = self.omu.app.id / str(rnd)
        return id.key()

    async def request(
        self,
        method: str,
        redirect: RequestRedirect,
        input: str | URL,
        headers: dict[str, str],
        body: bytes | io.BytesIO,
    ) -> HttpResponse:
        url = URL(input)
        id = self.generate_id()
        await self.omu.send(
            HTTP_REQUEST_CREATE,
            {
                "id": id,
                "header": headers,
                "method": method,
                "redirect": redirect,
                "url": url.human_repr(),
            },
        )
        if body:
            if isinstance(body, bytes):
                await self.omu.send(HTTP_REQUEST_SEND, DataChunk[RequestHandle]({"id": id}, body))
            else:
                while True:
                    # Read in 16MB chunks
                    chunk = body.read(1024 * 1024 * 16)
                    if not chunk:
                        break
                    await self.omu.send(HTTP_REQUEST_SEND, DataChunk[RequestHandle]({"id": id}, chunk))

        await self.omu.send(HTTP_REQUEST_CLOSE, {"id": id})
        chunks: bytes = b""
        responseFuture = Future[HttpResponse]()
        chunkEvent = Future[bool]()

        def receive(data: bytes):
            nonlocal chunks
            chunks += data
            nonlocal chunkEvent
            chunkEvent.set_result(True)
            chunkEvent = Future[bool]()

        def setResponse(response: HttpResponse):
            responseFuture.set_result(response)

            self.handles[id] = {
                "type": "receiving",
                "receive": receive,
                "close": lambda: chunkEvent.set_result(False),
            }

        self.handles[id] = {
            "type": "created",
            "setResponse": setResponse,
        }
        response = await responseFuture
        while await chunkEvent:
            ...
        return response



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from importlib.metadata import Distribution
from typing import TYPE_CHECKING, Protocol

from omu.omu import Omu

if TYPE_CHECKING:
    from omuserver.server import Server

from omu.helper import AsyncCallback


class PluginContext(Protocol):
    server: Server


class InstallContext(PluginContext):
    def __init__(
        self,
        server: Server,
        old_distribution: Distribution | None = None,
        new_distribution: Distribution | None = None,
        old_plugin: Plugin | None = None,
    ):
        self.server: Server = server
        self.old_distribution: Distribution | None = old_distribution
        self.new_distribution: Distribution | None = new_distribution
        self.plugin: Plugin | None = old_plugin
        self.restart_required: bool = False

    def require_restart(self):
        self.restart_required = True


@dataclass(frozen=True, slots=True)
class UninstallContext(PluginContext):
    server: Server
    distribution: Distribution | None
    plugin: Plugin | None


@dataclass(frozen=True, slots=True)
class StartContext(PluginContext):
    server: Server


@dataclass(frozen=True, slots=True)
class StopContext(PluginContext):
    server: Server


@dataclass(frozen=True, slots=True)
class Plugin:
    get_client: Callable[[], Omu] | None = None
    on_start: AsyncCallback[[StartContext]] | None = None
    on_stop: AsyncCallback[[StopContext]] | None = None
    on_install: AsyncCallback[[InstallContext]] | None = None
    on_uninstall: AsyncCallback[[UninstallContext]] | None = None
    on_update: AsyncCallback[[InstallContext]] | None = None
    isolated: bool = True

    def __post_init__(self):
        if self.isolated:
            assert self.get_client is not None, "Isolated plugins must have get_client"



from __future__ import annotations

from dataclasses import dataclass

import aiohttp
from aiohttp import ClientWebSocketResponse, web
from loguru import logger

from omu.address import Address
from omu.bytebuffer import ByteReader, ByteWriter
from omu.result import Err, Ok, Result
from omu.serializer import Serializable

from .connection import Connection, ConnectionClosed, ErrorReceiving, InvalidPacket, ReceiveError, Transport
from .packet import Packet, PacketData


@dataclass(frozen=True, slots=True)
class WebsocketsTransport(Transport):
    address: Address

    @property
    def _ws_endpoint(self) -> str:
        protocol = "wss" if self.address.secure else "ws"
        host = self.address.host or "127.0.0.1"
        port = self.address.port
        return f"{protocol}://{host}:{port}/ws"

    async def connect(self) -> WebsocketsConnection:
        session = aiohttp.ClientSession()
        socket = await session.ws_connect(self._ws_endpoint)
        return WebsocketsConnection(socket)


@dataclass(frozen=True, slots=True)
class WebsocketsConnection(Connection):
    socket: ClientWebSocketResponse

    async def send(self, packet: Packet, packet_mapper: Serializable[Packet, PacketData]) -> None:
        if not self.socket or self.socket.closed:
            raise RuntimeError("Not connected")
        packet_data = packet_mapper.serialize(packet)
        writer = ByteWriter()
        writer.write_string(packet_data.type)
        writer.write_uint8_array(packet_data.data)
        await self.socket.send_bytes(writer.finish())

    async def receive(self, packet_mapper: Serializable[Packet, PacketData]) -> Result[Packet, ReceiveError]:
        if not self.socket or self.socket.closed:
            return Err(ConnectionClosed("Socket is closed"))
        msg = await self.socket.receive()
        if msg.type in {
            web.WSMsgType.CLOSE,
            web.WSMsgType.CLOSED,
            web.WSMsgType.CLOSING,
            web.WSMsgType.ERROR,
        }:
            return Err(ConnectionClosed("Socket is closed"))
        if msg.data is None:
            return Err(ErrorReceiving("Received empty message"))
        if msg.type == web.WSMsgType.TEXT:
            return Err(ErrorReceiving("Received text message"))
        with ByteReader(msg.data) as reader:
            event_type = reader.read_string()
            event_data = reader.read_uint8_array()
        packet_data = PacketData(event_type, event_data)
        try:
            return Ok(packet_mapper.deserialize(packet_data))
        except Exception as err:
            logger.opt(exception=err).error("Failed to deserialize packet")
            return Err(InvalidPacket(f"Failed to deserialize packet: {packet_data.type}"))

    async def close(self) -> None:
        if not self.socket or self.socket.closed:
            return
        await self.socket.close()

    @property
    def closed(self) -> bool:
        return not self.socket or self.socket.closed



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Any, NotRequired, TypedDict

from omu.event_emitter import Unlisten
from omu.helper import Coro, map_optional
from omu.identifier import Identifier
from omu.serializer import Serializable, Serializer


class SignalPermissionsJSON(TypedDict):
    all: NotRequired[str | None]
    listen: NotRequired[str | None]
    notify: NotRequired[str | None]


@dataclass(frozen=True, slots=True)
class SignalPermissions:
    all: Identifier | None = None
    listen: Identifier | None = None
    notify: Identifier | None = None

    def to_json(self) -> SignalPermissionsJSON:
        return {
            "all": map_optional(self.all, Identifier.key),
            "listen": map_optional(self.listen, Identifier.key),
            "notify": map_optional(self.notify, Identifier.key),
        }

    @classmethod
    def from_json(cls, item: SignalPermissionsJSON) -> SignalPermissions:
        return SignalPermissions(
            all=map_optional(item.get("all"), Identifier.from_key),
            listen=map_optional(item.get("listen"), Identifier.from_key),
            notify=map_optional(item.get("notify"), Identifier.from_key),
        )


@dataclass(frozen=True, slots=True)
class SignalType[T]:
    id: Identifier
    serializer: Serializable[T, bytes]
    permissions: SignalPermissions = SignalPermissions()

    @classmethod
    def create_json(
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[T, Any] | None = None,
        permissions: SignalPermissions | None = None,
    ):
        return cls(
            id=identifier / name,
            serializer=Serializer.of(serializer or Serializer.noop()).to_json(),
            permissions=permissions or SignalPermissions(),
        )

    @classmethod
    def create_serialized(
        cls,
        identifier: Identifier,
        name: str,
        serializer: Serializable[T, bytes],
        permissions: SignalPermissions | None = None,
    ):
        return cls(
            id=identifier / name,
            serializer=serializer,
            permissions=permissions or SignalPermissions(),
        )


class Signal[T](abc.ABC):
    @abc.abstractmethod
    def listen(self, listener: Coro[[T], None]) -> Unlisten: ...

    @abc.abstractmethod
    async def notify(self, body: T) -> None: ...



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, LiteralString, TypedDict

from omu.api import Extension, ExtensionType
from omu.api.endpoint import EndpointType
from omu.api.permission.permission import PermissionTypeJson
from omu.api.plugin.package_info import PackageInfo
from omu.api.registry.registry import RegistryPermissions, RegistryType
from omu.api.server.extension import AppIndexRegistryMeta
from omu.api.table import TablePermissions, TableType
from omu.app import App, AppJson
from omu.bytebuffer import ByteReader, ByteWriter
from omu.errors import PermissionDenied
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.serializer import Serializer

from .packets import (
    AppInstallRequestPacket,
    AppInstallResponse,
    AppUpdateRequestPacket,
)

if TYPE_CHECKING:
    from omu.omu import Omu

DASHBOARD_EXTENSION_TYPE = ExtensionType("dashboard", lambda client: DashboardExtension(client))


class DashboardSetResponse(TypedDict):
    success: bool


DASHBOARD_SET_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "set"
DASHBOARD_SET_ENDPOINT = EndpointType[Identifier, DashboardSetResponse].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "set",
    request_serializer=Serializer.model(Identifier),
    permission_id=DASHBOARD_SET_PERMISSION_ID,
)
DASHBOARD_OPEN_APP_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "app" / "open"
DASHBOARD_OPEN_APP_ENDPOINT = EndpointType[App, None].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "open_app",
    request_serializer=Serializer.model(App),
    permission_id=DASHBOARD_OPEN_APP_PERMISSION_ID,
)
DASHBOARD_OPEN_APP_PACKET = PacketType[App].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "open_app",
    Serializer.model(App),
)
DASHBOARD_APP_INSTALL_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "app" / "install"
DASHBOARD_APP_INSTALL_ENDPOINT = EndpointType[App, AppInstallResponse].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "install_app",
    request_serializer=Serializer.model(App),
    permission_id=DASHBOARD_APP_INSTALL_PERMISSION_ID,
)
DASHBOARD_APP_INSTALL_PACKET = PacketType[AppInstallRequestPacket].create_serialized(
    DASHBOARD_EXTENSION_TYPE,
    "install_app",
    AppInstallRequestPacket,
)
DASHBOARD_APP_UPDATE_PACKET = PacketType[AppUpdateRequestPacket].create_serialized(
    DASHBOARD_EXTENSION_TYPE,
    "update_app",
    AppUpdateRequestPacket,
)

DASHBOARD_DRAG_DROP_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "drag_drop"


class DragDropFile(TypedDict):
    type: Literal["file", "directory"]
    size: int
    name: str


class DragDropPosition(TypedDict):
    x: int
    y: int


class DragEnter(TypedDict):
    type: Literal["enter"]
    drag_id: str
    files: list[DragDropFile]
    position: DragDropPosition


class DragOver(TypedDict):
    type: Literal["over"]
    drag_id: str
    position: DragDropPosition


class DragDrop(TypedDict):
    type: Literal["drop"]
    drag_id: str
    files: list[DragDropFile]
    position: DragDropPosition


class DragLeave(TypedDict):
    type: Literal["leave"]
    drag_id: str


type DragState = DragEnter | DragOver | DragDrop | DragLeave


class FileDragPacket(TypedDict):
    drag_id: str
    app: AppJson
    state: DragState


DASHBOARD_DRAG_DROP_STATE_PACKET = PacketType[FileDragPacket].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "drag_drop_state",
)


class DragDropReadRequest(TypedDict):
    drag_id: str


class DragDropReadRequestDashboard(TypedDict):
    request_id: str
    drag_id: str


class DragDropReadMeta(TypedDict):
    request_id: str
    drag_id: str
    files: list[DragDropFile]


class FileData(TypedDict):
    file: DragDropFile
    buffer: bytes


@dataclass
class DragDropReadResponse:
    meta: DragDropReadMeta
    files: dict[str, FileData]
    version: int = 1

    @staticmethod
    def serialize(item: DragDropReadResponse) -> bytes:
        writer = ByteWriter()
        writer.write_uleb128(item.version)
        writer.write_string(json.dumps(item.meta))
        writer.write_uleb128(len(item.files))

        for key, data in item.files.items():
            writer.write_string(key)
            writer.write_string(json.dumps(data["file"]))
            writer.write_uint8_array(data["buffer"])

        return writer.finish()

    @staticmethod
    def deserialize(item: bytes) -> DragDropReadResponse:
        with ByteReader(item) as reader:
            version = reader.read_uleb128()
            meta: DragDropReadMeta = json.loads(reader.read_string())
            length = reader.read_uleb128()

            files: dict[str, FileData] = {}
            for _ in range(length):
                key = reader.read_string()
                file: DragDropFile = json.loads(reader.read_string())
                buffer = reader.read_uint8_array()
                files[key] = {
                    "file": file,
                    "buffer": buffer,
                }

        return DragDropReadResponse(
            version=version,
            meta=meta,
            files=files,
        )


DASHBOARD_DRAG_DROP_READ_ENDPOINT = EndpointType[DragDropReadRequest, DragDropReadResponse].create_serialized(
    DASHBOARD_EXTENSION_TYPE,
    "drag_drop_read",
    request_serializer=Serializer.json(),
    response_serializer=DragDropReadResponse,
    permission_id=DASHBOARD_DRAG_DROP_PERMISSION_ID,
)

DASHBOARD_DRAG_DROP_READ_REQUEST_PACKET = PacketType[DragDropReadRequestDashboard].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "drag_drop_read_request",
)

DASHBOARD_DRAG_DROP_READ_RESPONSE_PACKET = PacketType[DragDropReadResponse].create_serialized(
    DASHBOARD_EXTENSION_TYPE,
    "drag_drop_read_response",
    serializer=DragDropReadResponse,
)


class DragDropRequest(TypedDict): ...


class DragDropRequestDashboard(TypedDict):
    request_id: str
    app: AppJson


class DragDropRequestResponse(TypedDict):
    request_id: str
    ok: bool


DASHBOARD_DRAG_DROP_REQUEST_ENDPOINT = EndpointType[DragDropRequest, DragDropRequestResponse].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "drag_drop_request",
    permission_id=DASHBOARD_DRAG_DROP_PERMISSION_ID,
)

DASHBOARD_DRAG_DROP_REQUEST_PACKET = PacketType[DragDropRequestDashboard].create_json(
    DASHBOARD_EXTENSION_TYPE, "drag_drop_request"
)


class PromptRequestBase[T: LiteralString](TypedDict):
    kind: T
    id: str


class PromptRequestAppPermissions(PromptRequestBase[Literal["app/permissions"]]):
    app: AppJson
    permissions: list[PermissionTypeJson]


class PromptRequestAppPlugins(PromptRequestBase[Literal["app/plugins"]]):
    app: AppJson
    packages: list[PackageInfo]


class PromptRequestAppInstall(PromptRequestBase[Literal["app/install"]]):
    app: AppJson
    dependencies: dict[str, AppJson]


class PromptRequestAppUpdate(PromptRequestBase[Literal["app/update"]]):
    old_app: AppJson
    new_app: AppJson


class PromptRequestIndexInstall(PromptRequestBase[Literal["index/install"]]):
    index_url: str
    meta: AppIndexRegistryMeta


type PromptRequest = (
    PromptRequestAppPermissions
    | PromptRequestAppPlugins
    | PromptRequestAppInstall
    | PromptRequestAppUpdate
    | PromptRequestIndexInstall
)

DASHBOARD_PROMPT_REQUEST = PacketType[PromptRequest].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "prompt_request",
)

type PromptResult = Literal["accept", "deny", "block"]


class PromptResponse(TypedDict):
    id: str
    kind: str
    result: PromptResult


DASHBOARD_PROMPT_RESPONSE = PacketType[PromptResponse].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "prompt_response",
)

DASHBOARD_PROMPT_CLEAR_BLOCKED = EndpointType[None, None].create_json(
    DASHBOARD_EXTENSION_TYPE,
    "prompt_clear_blocked",
    permission_id=DASHBOARD_SET_PERMISSION_ID,
)

DASHBOARD_WEBVIEW_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE / "webview"


@dataclass
class WebviewEventPacket:
    id: Identifier
    target: Identifier
    event: Any

    @staticmethod
    def serialize(packet: WebviewEventPacket):
        return {
            "id": packet.id.key(),
            "target": packet.target.key(),
            "event": packet.event,
        }

    @staticmethod
    def deserialize(packet: dict):
        return WebviewEventPacket(
            id=Identifier.from_key(packet["id"]),
            target=Identifier.from_key(packet["target"]),
            event=packet["event"],
        )


DASHBOARD_WEBVIEW_EVENT_PACKET = PacketType[WebviewEventPacket].create_json(
    DASHBOARD_EXTENSION_TYPE,
    name="webview_event",
    serializer=WebviewEventPacket,
)


class AllowedHost(TypedDict):
    host: str


DASHBOARD_ALLOWED_HOSTS = TableType[AllowedHost].create_json(
    DASHBOARD_EXTENSION_TYPE,
    name="hosts",
    key=lambda entry: entry["host"],
    permissions=TablePermissions(DASHBOARD_SET_PERMISSION_ID),
)

DASHBOARD_SPEECH_RECOGNITION_PERMISSION_ID = DASHBOARD_EXTENSION_TYPE.join("speech_recognition")


class TranscriptSegment(TypedDict):
    confidence: float
    transcript: str


class TranscriptStatusIdle(TypedDict):
    type: Literal["idle"]


class TranscriptStatusResult(TypedDict):
    type: Literal["result"]
    timestamp: float
    segments: list[TranscriptSegment]


class TranscriptStatusFinal(TypedDict):
    type: Literal["final"]
    timestamp: float
    segments: list[TranscriptSegment]


class TranscriptStatusAudioStarted(TypedDict):
    type: Literal["audio_started"]
    timestamp: float


class TranscriptStatusAudioEnded(TypedDict):
    type: Literal["audio_ended"]
    timestamp: float


type TranscriptStatus = (
    TranscriptStatusIdle
    | TranscriptStatusResult
    | TranscriptStatusFinal
    | TranscriptStatusAudioStarted
    | TranscriptStatusAudioEnded
)


DASHBOARD_SPEECH_RECOGNITION = RegistryType[TranscriptStatus].create_json(
    DASHBOARD_EXTENSION_TYPE,
    name="speech_recognition",
    default_value={"type": "idle"},
    permissions=RegistryPermissions(
        read=DASHBOARD_SPEECH_RECOGNITION_PERMISSION_ID,
        write=DASHBOARD_SET_PERMISSION_ID,
    ),
)


class SpeechRecognitionStart(TypedDict):
    type: Literal["start"]


DASHBOARD_SPEECH_RECOGNITION_START = EndpointType[SpeechRecognitionStart, Any].create_json(
    DASHBOARD_EXTENSION_TYPE,
    name="speech_recognition_start",
    permission_id=DASHBOARD_SPEECH_RECOGNITION_PERMISSION_ID,
)


class DashboardExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return DASHBOARD_EXTENSION_TYPE

    def __init__(self, omu: Omu):
        self.omu = omu
        self.omu.network.register_packet(
            DASHBOARD_OPEN_APP_PACKET,
            DASHBOARD_APP_INSTALL_PACKET,
            DASHBOARD_APP_UPDATE_PACKET,
            DASHBOARD_DRAG_DROP_STATE_PACKET,
            DASHBOARD_DRAG_DROP_READ_REQUEST_PACKET,
            DASHBOARD_DRAG_DROP_READ_RESPONSE_PACKET,
            DASHBOARD_DRAG_DROP_REQUEST_PACKET,
        )

    async def open_app(self, app: App) -> None:
        if not self.omu.permissions.has(DASHBOARD_OPEN_APP_PERMISSION_ID):
            error = f"Pemission {DASHBOARD_OPEN_APP_PERMISSION_ID} required to open app {app}"
            raise PermissionDenied(error)
        await self.omu.endpoints.call(DASHBOARD_OPEN_APP_ENDPOINT, app)



from __future__ import annotations

import base64
import secrets
from dataclasses import dataclass

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import padding as aes_padding
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from omu.bytebuffer import ByteReader, ByteWriter
from omu.network.packet.packet import Packet, PacketData
from omu.network.packet.packet_types import PACKET_TYPES, RSANumbers

def urlsafe_b64encode(data: bytes) -> str:
    return base64.urlsafe_b64encode(data).decode().rstrip("=")

def urlsafe_b64decode(data: str) -> bytes:
    padding_needed = 4 - (len(data) % 4)
    if padding_needed and padding_needed != 4:
        data += "=" * padding_needed
    return base64.urlsafe_b64decode(data)

@dataclass(frozen=True, slots=True)
class Decryptor:
    private: rsa.RSAPrivateKey
    public: rsa.RSAPublicKey

    @classmethod
    def new(cls) -> Decryptor:
        private = rsa.generate_private_key(
            public_exponent=0x010001,
            key_size=4096,
        )
        public = private.public_key()
        return Decryptor(
            private=private,
            public=public,
        )

    def to_request(self) -> RSANumbers:
        numbers = self.public.public_numbers()
        return {
            "e": urlsafe_b64encode(numbers.e.to_bytes((0x010001.bit_length() + 7) // 8, "big")),
            "n": urlsafe_b64encode(numbers.n.to_bytes((numbers.n.bit_length() + 7) // 8, "big")),
        }

    def decrypt(self, data: bytes) -> bytes:
        decrypted = self.private.decrypt(
            data,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        return decrypted

    def decrypt_string(self, data: str) -> str:
        return self.decrypt(urlsafe_b64decode(data)).decode()


@dataclass(frozen=True, slots=True)
class Encryptor:
    public: rsa.RSAPublicKey

    @classmethod
    def new(cls, request: RSANumbers) -> Encryptor:
        public_numbers = rsa.RSAPublicNumbers(
            e=int.from_bytes(urlsafe_b64decode(request["e"])),
            n=int.from_bytes(urlsafe_b64decode(request["n"])),
        )
        return Encryptor(
            public=public_numbers.public_key(),
        )

    def encrypt(self, data: bytes) -> bytes:
        encrypted = self.public.encrypt(
            data,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        return encrypted

    def encrypt_string(self, data: str) -> str:
        return urlsafe_b64encode(self.encrypt(data.encode()))


@dataclass(slots=True)
class AES:
    key: bytes
    iv: bytes
    cipher: Cipher
    counter: int = 0
    pad = aes_padding.PKCS7(128)

    @classmethod
    def new(cls) -> AES:
        key = secrets.token_bytes(32)
        iv = secrets.token_bytes(16)
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
        return AES(key=key, iv=iv, cipher=cipher)

    @classmethod
    def deserialize(cls, aes: str, decryptor: Decryptor) -> AES:
        decrypted = decryptor.decrypt(urlsafe_b64decode(aes))
        with ByteReader(decrypted) as reader:
            key = reader.read_uint8_array()
            iv = reader.read_uint8_array()
        return AES(
            key=key,
            iv=iv,
            cipher=Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend()),
        )

    def serialize(self, encryptor: Encryptor) -> str:
        writer = ByteWriter()
        writer.write_uint8_array(self.key)
        writer.write_uint8_array(self.iv)
        return urlsafe_b64encode(encryptor.encrypt(writer.finish()))

    def encrypt(self, packet_data: PacketData) -> Packet:
        writer = ByteWriter()
        writer.write_string(packet_data.type)
        writer.write_uint8_array(packet_data.data)

        padder = self.pad.padder()
        padded_data = padder.update(writer.finish()) + padder.finalize()
        encryptor = self.cipher.encryptor()
        encrypted = encryptor.update(padded_data) + encryptor.finalize()
        return Packet(PACKET_TYPES.ENCRYPTED_PACKET, encrypted)

    def decrypt(self, packet: Packet[bytes]) -> PacketData:
        decryptor = self.cipher.decryptor()
        decrypted = decryptor.update(packet.data) + decryptor.finalize()
        unpadder = self.pad.unpadder()
        unpadded_data = unpadder.update(decrypted) + unpadder.finalize()

        with ByteReader(unpadded_data) as reader:
            type = reader.read_string()
            data = reader.read_uint8_array()
            return PacketData(type=type, data=data)



from __future__ import annotations

from omu.api import Extension, ExtensionType
from omu.event_emitter import Unlisten
from omu.helper import Coro
from omu.identifier import Identifier
from omu.network.packet import PacketType
from omu.omu import Omu
from omu.serializer import Serializer

from .packets import SignalPacket, SignalRegisterPacket
from .signal import Signal, SignalType

SIGNAL_EXTENSION_TYPE = ExtensionType("signal", lambda client: SignalExtension(client))


SIGNAL_REGISTER_PACKET = PacketType[SignalRegisterPacket].create_json(
    SIGNAL_EXTENSION_TYPE,
    "register",
    SignalRegisterPacket,
)
SIGNAL_LISTEN_PACKET = PacketType[Identifier].create_json(
    SIGNAL_EXTENSION_TYPE,
    "listen",
    serializer=Serializer.model(Identifier),
)
SIGNAL_NOTIFY_PACKET = PacketType[SignalPacket].create_serialized(
    SIGNAL_EXTENSION_TYPE,
    "notify",
    SignalPacket,
)


class SignalExtension(Extension):
    @property
    def type(self) -> ExtensionType:
        return SIGNAL_EXTENSION_TYPE

    def __init__(self, omu: Omu):
        self.client = omu
        self.signals: dict[Identifier, Signal] = {}
        omu.network.register_packet(
            SIGNAL_REGISTER_PACKET,
            SIGNAL_LISTEN_PACKET,
            SIGNAL_NOTIFY_PACKET,
        )

    def create_signal[T](self, signal_type: SignalType[T]) -> Signal[T]:
        if signal_type.id in self.signals:
            raise Exception(f"Signal {signal_type.id} already exists")
        return SignalImpl(self.client, signal_type)

    def create[T](self, name: str, _t: type[T] | None = None) -> Signal[T]:
        identifier = self.client.app.id / name
        type = SignalType[T].create_json(
            identifier,
            name,
        )
        return self.create_signal(type)

    def get[T](self, signal_type: SignalType[T]) -> Signal[T]:
        return self.create_signal(signal_type)


class SignalImpl[T](Signal):
    def __init__(self, omu: Omu, type: SignalType[T]):
        self.client = omu
        self.type = type
        self.listeners: list[Coro[[T], None]] = []
        self.listening = False
        omu.network.add_packet_handler(SIGNAL_NOTIFY_PACKET, self._on_broadcast)
        omu.network.add_task(self._on_task)

    async def notify(self, body: T) -> None:
        data = self.type.serializer.serialize(body)
        await self.client.send(
            SIGNAL_NOTIFY_PACKET,
            SignalPacket(id=self.type.id, body=data),
        )
        for listener in self.listeners:
            await listener(body)

    def listen(self, listener: Coro[[T], None]) -> Unlisten:
        if not self.listening:

            async def on_ready():
                await self.client.send(SIGNAL_LISTEN_PACKET, self.type.id)

            self.client.on_ready(on_ready)
            self.listening = True

        self.listeners.append(listener)
        return lambda: self.listeners.remove(listener)

    async def _on_task(self) -> None:
        if not self.type.id.is_subpath_of(self.client.app.id):
            return
        packet = SignalRegisterPacket(
            id=self.type.id,
            permissions=self.type.permissions,
        )
        await self.client.send(
            SIGNAL_REGISTER_PACKET,
            packet,
        )

    async def _on_broadcast(self, data: SignalPacket) -> None:
        if data.id != self.type.id:
            return

        body = self.type.serializer.deserialize(data.body)
        for listener in self.listeners:
            await listener(body)
